""" DOS/Door wrapper for x/84. """
# std imports
import contextlib
import shlex
import os

# local
from x84.bbs import getsession, getterminal, echo, ini, Pager, Dropfile
from x84.bbs import Door, DOSDoor, get_ini, syncterm_setfont


def prompt_resize_term(session, term, name):
    want_cols = get_ini('sesame', '{0}_cols'.format(name), getter='getint')
    want_rows = get_ini('sesame', '{0}_rows'.format(name), getter='getint')

    if want_cols and want_rows and not term.kind.startswith('ansi'):
        while not (term.width == want_cols and
                   term.height == want_rows):
            resize_msg = (u'Please resize your window to {0} x {1} '
                          u'current size is {term.width} x {term.height} '
                          u'or press return.  Press escape to cancel.'
                          .format(want_cols, want_rows, term=term))
            echo(term.normal + term.home + term.clear_eos + term.home)
            pager = Pager(yloc=0, xloc=0, width=want_cols, height=want_rows,
                          colors={'border': term.bold_red})
            echo(term.move(term.height - 1, term.width))

            echo(pager.refresh() + pager.border())

            width = min(70, term.width - 5)
            for yoff, line in enumerate(term.wrap(resize_msg, width)):
                echo(u''.join((term.move(yoff + 1, 5), line.rstrip())))

            event, data = session.read_events(('input', 'refresh'))
            if event == 'refresh':
                continue

            if event == 'input':
                session.buffer_input(data, pushback=True)
                inp = term.inkey(0)
                while inp:
                    if inp.code == term.KEY_ENTER:
                        echo(term.normal + term.home + term.clear_eos)
                        return True
                    if inp.code == term.KEY_ESCAPE:
                        return False
                    inp = term.inkey(0)
        return True
    # if terminal type is 'ansi', just pass-through
    return True


def restore_screen(term, cols, rows):
    term._columns, term._rows = cols, rows


def parse_command_args(session, name, node):
    # Parse command path and arguments
    command = get_ini('sesame', name)
    if ' ' in command:
        command, args = command.split(' ', 1)
        args = args.format(
            session=session.to_dict(),
            system=dict(ini.CFG.items('system')),
            node=node,
        )
        args = shlex.split(args)
    else:
        args = []

    if not os.path.exists(command):
        raise RuntimeError("The door {0} specified a command path of "
                           "{1!r}, but no such file exists."
                           .format(name, command))

    return command, args


@contextlib.contextmanager
def acquire_node(session, name):
    nodes = get_ini('sesame', '{0}_nodes'.format(name), getter='getint')
    if not nodes:
        yield None
        return

    for node in range(1, nodes + 1):
        event = 'lock-{name}/{node}'.format(name=name, node=node)
        session.send_event(event, ('acquire', None))
        if session.read_event(event):
            yield node
            session.send_event(event, ('release', None))
            return

    # node could not be acquired
    yield -1


def get_env(session, name):
    # Setup the OS Environment values (if any)
    env = dict()
    env_prefix = '{0}_env_'.format(name)
    for option in ini.CFG.options('sesame'):
        if option.startswith(env_prefix):
            key = option.replace(env_prefix, '').upper()
            env[key] = ini.CFG.get('sesame', option).format(
                session=session.to_dict(),
                system=dict(ini.CFG.items('system')),
            )
    return env


def do_dropfile(name, node):
    dropfile_path = get_ini('sesame', '{0}_droppath'.format(name))
    dropfile_type = get_ini('sesame', '{0}_droptype'.format(name)
                            ) or 'doorsys'
    if not dropfile_path:
        return

    dropfile_type = dropfile_type.upper()

    if not hasattr(Dropfile, dropfile_type):
        raise ValueError('sesame configuration declares dropfile '
                         'format of {0!r} but value is not supported '
                         'by class Dropfile.'.format(dropfile_type))

    _dropfile_type = getattr(Dropfile, dropfile_type)
    Dropfile(_dropfile_type, node).save(dropfile_path)


def main(name):
    """ Sesame runs a named door. """
    session, term = getsession(), getterminal()

    # clear screen,
    echo(term.normal + u'\r\n' * term.height + term.move(0, 0))

    # set font,
    if term.kind.startswith('ansi'):
        syncterm_font = get_ini(
            'sesame', '{0}_syncterm_font'.format(name)) or 'cp437'
        echo(syncterm_setfont(syncterm_font))
        echo(term.move_x(0) + term.clear_eol)

    # pylint: disable=W0212
    #         Access to a protected member {_columns, _rows} of a client class
    store_columns, store_rows = term._columns, term._rows
    if not prompt_resize_term(session, term, name):
        return

    with acquire_node(session, name) as node:

        if node == -1:
            echo(term.bold_red('This door has reached the maximum number '
                               'of nodes and may not be played.\r\n\r\n'
                               'press any key.'))
            term.inkey()
            restore_screen(term, store_columns, store_rows)
            return

        do_dropfile(name, node)

        session.activity = u'Playing {}'.format(name)

        cmd, args = parse_command_args(session, name, node)
        env = get_env(session, name)
        cp437 = get_ini('sesame', '{0}_cp437'.format(name),
                        getter='getboolean')

        _Door = DOSDoor if cmd.endswith('dosemu') else Door
        _Door(cmd=cmd, args=args, env=env, cp437=cp437).run()

    restore_screen(term, store_columns, store_rows)



"""
Encoding selection script for x/84.

Displays a CP437 artwork (block ansi), and prompts the user
to chose 'utf8' or 'cp437' encoding.
"""
# std imports
import os

# x/84
from x84.bbs import getsession, getterminal, echo, syncterm_setfont, LineEditor

# local
from common import display_banner, coerce_terminal_encoding

#: filepath to folder containing this script
here = os.path.dirname(__file__)

#: filepath to artfile displayed for this script
art_file = os.path.join(here, 'art', 'encoding.ans')

#: encoding used to display artfile
art_encoding = 'cp437_art'

#: fontset for SyncTerm emulator
syncterm_font = 'cp437'

#: text to display in prompt
prompt_text = (u"Chose an encoding and set font until artwork "
               u"looks best.  "

               u"SyncTerm, netrunner, and other DOS-emulating clients "
               u"should chose cp437, though internationalized languages "
               u"will appear as '?'.  "

               u"OSX Clients should chose the Andale Mono font. Linux "
               u"clients should chose the Deja Vu Sans Mono font. ")

#: horizontal margin from window edge for prompt
prompt_padding = 10


def _show_opt(term, keys):
    """ Display characters ``key`` highlighted as keystroke """
    return u''.join((term.bold_black(u'['),
                     term.bold_green_underline(keys),
                     term.bold_black(u']')))


def display_prompt(term):
    """ Display prompt of user choices. """
    echo(u'\r\n\r\n')
    width = min(term.width, 80 - prompt_padding)
    for line in term.wrap(prompt_text, width):
        echo(u' ' * ((term.width - width) / 2))
        echo(line + '\r\n')

    echo(u'\r\n')
    echo(term.center(u'{0}tf8, {1}p437, {2}one:'
                     .format(_show_opt(term, u'u'),
                             _show_opt(term, u'c'),
                             _show_opt(term, u'd'))
                     ).rstrip() + u' ')


def do_select_encoding(term, session):
    editor_colors = {'highlight': term.black_on_green}
    dirty = True
    while True:
        if session.poll_event('refresh') or dirty:
            # attempt to coerce encoding of terminal to match session.
            coerce_terminal_encoding(term, session.encoding)

            # display artwork and prompt
            vertical_padding = 2 if term.height >= 24 else 0
            display_banner(filepattern=art_file,
                           encoding=art_encoding,
                           vertical_padding=vertical_padding)
            display_prompt(term)
            dirty = False

        inp = LineEditor(1, colors=editor_colors).read()

        if inp is None or inp.lower() == 'd':
            break
        elif len(inp):
            # bad input -- reposition cursor for next LineEditor()
            echo(u'\b')
        if inp.lower() == u'u' and session.encoding != 'utf8':
            # switch to utf8,
            session.encoding = 'utf8'
            dirty = True
        elif inp.lower() == 'c' and session.encoding != 'cp437':
            # or cp437
            session.encoding = 'cp437'
            dirty = True


def main():
    """ Script entry point. """
    session, term = getsession(), getterminal()
    session.activity = u'Selecting character set'

    # set syncterm font, if any
    if syncterm_font and term.kind.startswith('ansi'):
        echo(syncterm_setfont(syncterm_font))

    do_select_encoding(term, session)



""" Module provides ``Script`` definition for x/84. """
# std imports
import collections

#: Defines a target script name, positional, and keyword arguments.
Script = collections.namedtuple('Script', ['name', 'args', 'kwargs'])



""" Chat script for x/84. """
# std imports
from __future__ import division
import collections
import logging
import math

# local
from x84.bbs import echo, getsession, getterminal, syncterm_setfont, showart
from x84.bbs import ScrollingEditor, get_ini
from x84.bbs.session import Script

ChatEvent = collections.namedtuple('ChatEvent', [
    'session_id', 'channel', 'handle', 'command', 'cmd_args'])

WindowDimension = collections.namedtuple(
    'WindowDimension', ['yloc', 'xloc', 'height', 'width'])

#: for ytalk-like version of split-screen of small windows, show user@bbs
system_bbsname = get_ini(section='system', key='bbsname')

ANSWER, HANGUP, REJECTED = 1, 2, 3

art_encoding = 'cp437'

#: filepath to artfile displayed for this script
art_file = get_ini(
    section='chat', key='art_file'
) or 'art/chat.ans'

#: encoding used to display artfile
art_encoding = get_ini(
    section='chat', key='art_encoding'
) or 'cp437'

#: fontset for SyncTerm emulator
syncterm_font = get_ini(
    section='chat', key='syncterm_font'
) or 'cp437'


def prompt_chat_request(term, pos, call_from):
    chat_prompt = (
        u'{marker}{nick} would like to chat, accept? {lb}{yn}{rb}'
        .format(marker=term.bold(u' ** '),
                nick=term.bold(call_from),
                lb=term.bold_black(u'['),
                rb=term.bold_black(u']'),
                yn=term.bold(u'yn')))
    echo(term.normal + u'\a')
    echo(term.move(pos.yloc, pos.xloc))
    echo(chat_prompt.center(pos.width, ' '))
    while True:
        inp = term.inkey()
        if inp.lower() == u'y':
            return True
        elif inp.lower() == u'n':
            return False


def refresh_screen(term, who):
    # create a new, empty screen
    echo(term.normal)
    echo(term.move(term.height - 1, 0))
    echo(u'\r\n' * (term.height + 1))

    ans_height, ans_width = 24, 80
    if term.width < ans_width or term.height < ans_height:
        # screen is too small for ansi file, use ytalk-like format
        center_line = term.height // 2
        echo(term.move(center_line, 0))
        echo(term.center(u'= {0}@{1} ='.format(who, system_bbsname),
                         term.width, '-'))
        top_window = WindowDimension(
            yloc=-1, xloc=-1,
            height=center_line + 2,
            width=term.width + 2)
        bot_window = WindowDimension(
            yloc=center_line,
            xloc=-1,
            height=term.height - (center_line - 1),
            width=term.width + 2)
    else:
        # center ansi file
        center_y = (math.floor(term.height / 2 - (ans_height / 2)))
        center_x = math.floor(term.width / 2 - (ans_width / 2))
        ypos, xpos = max(0, int(center_y)), max(0, int(center_x))
        art_generator = showart(art_file, encoding=art_encoding)
        for y_offset, line in enumerate(art_generator):
            _ypos = ypos + y_offset
            if _ypos <= term.height:
                echo(term.move(_ypos, xpos))
                echo(line.rstrip())
        top_window = WindowDimension(
            yloc=ypos,
            xloc=xpos + 1,
            height=9, width=78)
        bot_window = WindowDimension(
            yloc=ypos + 14,
            xloc=xpos + 1,
            height=10, width=78)
    return top_window, bot_window


def make_editor_series(winsize, editors):
    colors = {'highlight': u''}
    # prevent horizontal scrolling prior to final column
    margin_pct = 100 - (100 / (winsize.width - 3))
    # scroll at final column
    scroll_pct = 100 / (winsize.width - 3)
    return [
        ScrollingEditor(
            yloc=_yloc, xloc=winsize.xloc, width=winsize.width,
            colors=colors, margin_pct=margin_pct, scroll_pct=scroll_pct,
            content=(u'' if not editors or _idx >= len(editors)
                     else editors[_idx].content),
            max_length=winsize.width
        ) for _idx, _yloc in enumerate(range(
            winsize.yloc, winsize.yloc + winsize.height + -2))]


def get_editors(top_winsize, bot_winsize, top_editors, bot_editors):
    return (make_editor_series(top_winsize, top_editors),
            make_editor_series(bot_winsize, bot_editors))


def do_chat(session, term, log, other_sid, dial=None, call_from=None):
    """ Main procedure. """
    editor = None
    dirty = True
    top_idx = bot_idx = 0
    dialing = bool(dial)
    answering = bool(call_from)

    data = session.flush_event('chat')
    if data:
        log.debug('Flushed chat data: %r', data)

    session.activity = (u'{doing} {call_from}'.format(
        doing=('answering call from' if answering else
               'requesting chat with'),
        call_from=call_from))

    # send chat request
    if dial:
        # force other user to gosub this script, with 'other_sid' as ours.
        script = Script(name='chat', args=(),
                        kwargs={'call_from': session.user.handle,
                                'other_sid': session.sid})
        event, data = 'gosub', script
        route_data = (other_sid, event) + tuple(data)
        session.send_event('route', route_data)

    top_editors, bot_editors = [], []
    while True:
        if dirty:
            who = dial if dialing else call_from
            top_winsize, bot_winsize = refresh_screen(term, who)
            top_editors, bot_editors = get_editors(top_winsize, bot_winsize,
                                                   top_editors, bot_editors)

            # when a screen is resized, we must adjust our editor indicies to
            # ensure they are within range
            top_idx = min(top_idx, len(top_editors) - 1)
            bot_idx = min(bot_idx, len(bot_editors) - 1)

            for editor in top_editors + bot_editors:
                echo(editor.clear())
                echo(editor.refresh())

            echo(top_editors[top_idx].fixate())
            if dialing:
                display_dialing(term, pos=bot_winsize, who=who)
            if answering:
                display_answering(term, pos=bot_winsize, who=who)
            dirty = False

        event, data = session.read_events(
            events=('chat', 'input', 'refresh', 'chat'))

        if event == 'refresh':
            dirty = True
        elif event == 'input':
            session.buffer_input(data, pushback=True)
            if answering:
                if do_answer(term, session, other_sid) is False:
                    # 'n', hangup
                    break
                answering = False
                dirty = True
            elif dialing:
                inp = term.inkey(0)
                if inp.lower() == u'q' or inp.code == term.KEY_ESCAPE:
                    break
            else:
                # process keystroke and send to other party
                top_editors, top_idx = do_input(
                    term, session, editors=top_editors, edit_idx=top_idx,
                    other_sid=other_sid)
                if top_editors is None and top_idx is None:
                    # do_input() returns (None, None) on exit
                    break

        elif event == 'chat':
            if data == (HANGUP,):
                display_hangup(term, pos=top_winsize, who=call_from or dial)
                term.inkey()
                break
            elif data == (REJECTED,):
                display_rejected(term, pos=bot_winsize, who=dial)
                term.inkey()
                break
            elif dialing and data == (ANSWER,):
                dialing = False
                dirty = True
                continue
            elif isinstance(data[0], basestring):
                bot_editors, bot_idx = recv_input(editors=bot_editors,
                                                  edit_idx=bot_idx,
                                                  inp=data[0])
            else:
                log.error("Unexpected data for event 'chat': {0!r}", data)


def display_dialing(term, pos, who):
    echo(term.move(pos.yloc + (pos.height // 2), pos.xloc))
    echo(u'dialing {0}, [q]uit ...'.format(who).center(pos.width).rstrip())


def display_answering(term, pos, who):
    echo(term.move(pos.yloc + (pos.height // 2), pos.xloc))
    echo(u'accept chat from {0} [yn] ?'.format(who).center(pos.width).rstrip())
    echo(u'\b\b')


def display_hangup(term, pos, who):
    echo(term.move(pos.yloc + (pos.height // 2), pos.xloc))
    echo(u'{0} hung up.'.format(who).center(pos.width).rstrip())


def display_rejected(term, pos, who):
    echo(term.move(pos.yloc + (pos.height // 2), pos.xloc))
    echo(u'{0} is rejecting chat requests.'.format(who).center(pos.width).rstrip())


def do_answer(term, session, other_sid):
    while True:
        inp = term.inkey()
        if inp.lower() == u'y':
            route_data = (other_sid, 'chat') + (ANSWER,)
            session.send_event('route', route_data)
            return True
        elif inp.lower() == u'n':
            return False


def clear_editor(editor):
    """ Clear content and refresh line editor ``editor``. """
    editor.update(u'')
    echo(editor.refresh())


def recv_input(editors, edit_idx, inp):
    nextline = False
    editor = editors[edit_idx]

    # position cursor
    echo(editor.fixate())

    if inp.is_sequence:
        if ((inp.code in editor.keyset['backspace'] or
             inp in editor.keyset['backspace']
             ) and editor.content == u''):
            # editor is empty, and we've backspaced, allow
            # backspacing into previous line
            edit_idx = (len(editors) - 1 if edit_idx == 0
                        else edit_idx - 1)
            editor = editors[edit_idx]
            editor.update(editor.content)
            echo(editor.refresh() + editor.fixate())
        echo(editor.process_keystroke(inp))
    else:
        echo(editor.process_keystroke(inp))

    # return pressed or at margin (bell)
    nextline = (editor.carriage_returned or
                editor.bell)

    if nextline:
        edit_idx = (0 if edit_idx == len(editors) - 1
                    else edit_idx + 1)

        # clear new editor-line
        clear_editor(editors[edit_idx])
        if edit_idx + 1 < len(editors):
            # as well as the next
            clear_editor(editors[edit_idx + 1])
        echo(editors[edit_idx].fixate())

    return editors, edit_idx


def do_input(term, session, editors, edit_idx, other_sid):
    inp = term.inkey(0)
    while inp:
        if inp.is_sequence and inp.code == term.KEY_ESCAPE:
            # escape was pressed, exit
            return (None, None)
        editors, edit_idx = recv_input(editors, edit_idx, inp)

        route_data = (other_sid, 'chat') + (inp,)
        session.send_event('route', route_data)

        inp = term.inkey(0)
    return editors, edit_idx


def do_hangup(session, other_sid):
    route_data = (other_sid, 'chat') + (HANGUP,)
    session.send_event('route', route_data)


def main(*args, **kwargs):
    session, term = getsession(), getterminal()
    log = logging.getLogger(__name__)

    if kwargs.get('call_from') and session.kind not in ('telnet', 'ssh'):
        # reject chat requests unless we are a tty terminal of telnet or ssh.
        route_data = (kwargs['other_sid'], 'chat') + (REJECTED,)
        session.send_event('route', route_data)
        log.debug("Rejected chat request; session.kind=%s", session.kind)
        return True

    # set syncterm font, if any
    if syncterm_font and term.kind.startswith('ansi'):
        echo(syncterm_setfont(syncterm_font))

    with term.fullscreen():
        try:
            return do_chat(session, term, log=log, *args, **kwargs)
        finally:
            do_hangup(session, other_sid=kwargs['other_sid'])



# -*- coding: utf-8 -*-
"""
New user account script for x/84.

This script is closely coupled with, and dependend on by profile.py.
"""
# std imports
from __future__ import division
import collections
import logging
import os
import re

# local
from x84.bbs import getsession, getterminal, echo, LineEditor
from x84.bbs import goto, get_ini, find_user, User, syncterm_setfont
from common import display_banner, show_description

log = logging.getLogger(__name__)
here = os.path.dirname(__file__)

#: banner art displayed in main()
art_file = get_ini(
    section='nua', key='art_file'
) or os.path.join(here, 'art', 'nua*.ans')

#: encoding of artfile
art_encoding = get_ini(
    section='nua', key='art_encoding'
) or 'ascii'

#: preferred fontset for SyncTerm emulator
syncterm_font = get_ini(
    section='nua', key='syncterm_font'
) or 'topaz'

#: which script to execute after registration
top_script = get_ini(
    section='matrix', key='topscript'
) or 'top'

#: which login names trigger new user account script
new_usernames = get_ini(
    section='matrix', key='newcmds', split=True
) or ['new']

#: maximum length of user handles
username_max_length = get_ini(
    section='nua', key='max_user', getter='getint'
) or 10

#: minimum length of user handles
username_min_length = get_ini(
    section='nua', key='min_user', getter='getint',
) or 2

#: which login names trigger new user account script
invalid_usernames = get_ini(
    section='nua', key='invalid_handles', split=True
) or ['new', 'apply', 'exit', 'logoff', 'bye', 'quit', 'sysop', 'anonymous'
      ] + new_usernames

#: login name validation as a regular expression
username_re_validator = get_ini(
    section='nua', key='handle_validation'
) or ['^[A-Za-z0-9]{3,11}$']


#: maximum length of user 'location' field
location_max_length = get_ini(
    section='nua', key='max_location', getter='getint'
) or 25

#: maximum length of email address
email_max_length = get_ini(
    section='nua', key='max_email', getter='getint'
) or 30

#: maximum length of password
password_max_length = get_ini(
    section='nua', key='max_pass', getter='getint'
) or 15

#: minimum length of password
password_min_length = get_ini(
    section='nua', key='min_pass', getter='getint'
) or 4

#: primary color (highlight)
color_primary = get_ini(
    section='nua', key='color_primary'
) or 'magenta'

#: secondary color (lowlight)
color_secondary = get_ini(
    section='nua', key='color_secondary'
) or 'red'

#: password hidden character
hidden_char = get_ini(
    section='nua', key='hidden_char'
) or u'\u00f7'

#: structure for prompting input/validation
vfield = collections.namedtuple('input_validation', [
    # field name of user record
    'name',
    # field query displayed to user
    'prompt_key',
    # dictionary of arguments passed to LineEditor
    'kwargs',
    # validation function, None if no validation is required.
    # function receives (user, new_value) and returns tuple
    # of (errmsg, index modifier).
    'validation_function',
    # when resuming input, how to retrieve field value, may
    # return None (for example, the password field)
    'getter',
    # describe the necessity or optionality of the field and
    # its purpose to the user
    'description',
])


def validate_handle(_, handle):
    """ Validate user ``handle``. """
    errmsg = None
    if find_user(handle):
        errmsg = u'User by this name already exists.'

    elif len(handle) < username_min_length:
        errmsg = (u'Username too short, must be at least {0} characters.'
                  .format(username_min_length))

    elif handle.lower() in invalid_usernames:
        errmsg = u'Username is not legal form: reserved.'

    elif os.path.sep in handle:
        errmsg = u'Username is not legal form: contains OS path separator.'

    elif not re.match(username_re_validator, handle):
        errmsg = (u'Username fails validation of regular expression, '
                  u'{0!r}.'.format(username_re_validator))

    # No validation error
    return errmsg, 0


def validate_password(_, password):
    """ Validate setting ``password`` for ``user``. """
    errmsg = None
    if password == u'':
        errmsg = u'Password is required.'
    elif len(password) < password_min_length:
        errmsg = (u'Password too short, must be at least {0} characters.'
                  .format(password_min_length))

    # No validation error
    return errmsg, 0


def validate_password_again(user, password):
    """ Validate 2nd round of ``password`` for ``user``. """
    errmsg = None
    if password == u'':
        errmsg = u'Password is required.'
    elif len(password) < password_min_length:
        errmsg = (u'Password too short, must be at least {0} characters.'
                  .format(password_min_length))
    elif not user.auth(password):
        errmsg = u'Password does not match, try again!'

    # returns -1 as modifier, to return to password prompt.
    return errmsg, -1


def get_validation_fields(user):
    """ Return field validation definitions prompted by do_nua() loop. """
    fields = collections.OrderedDict()
    fields['handle'] = vfield(
        name='handle',
        prompt_key='Username',
        kwargs={'width': username_max_length},
        validation_function=validate_handle,
        getter=lambda: getattr(user, 'handle', None),
        description=(u'Handle or Alias you will be known '
                     u'by on this board.'))
    fields['location'] = vfield(
        name='location',
        prompt_key='Origin (optional)',
        kwargs={'width': location_max_length},
        validation_function=None,
        getter=lambda: getattr(user, 'location', None),
        description=(u'Group of affiliation, geographic location, '
                     u'or other moniker which other members will '
                     u'know as your place or origin.'))
    fields['email'] = vfield(
        name='email',
        prompt_key='E-mail (optional)',
        kwargs={'width': email_max_length},
        validation_function=None,
        getter=lambda: getattr(user, 'email', None),
        description=(u'E-mail address is both private and optional, '
                     u'allowing the ability to reset your password '
                     u'should it ever be forgotten.'))
    fields['password'] = vfield(
        name='password',
        prompt_key='Password',
        kwargs={'width': password_max_length,
                'hidden': hidden_char},
        validation_function=validate_password,
        getter=lambda: None,
        description=None)
    fields['password_again'] = vfield(
        name=None,
        prompt_key='Again',
        kwargs={'width': password_max_length,
                'hidden': hidden_char},
        validation_function=validate_password_again,
        getter=lambda: None,
        description=None)
    return fields


def fixate_next(term, newlines=2):
    """ positions next prompt 40 pixels minus center of screen. """
    return u''.join((
        u'\r\n\r\n' * newlines,
        term.move_x(max(0, (term.width // 2) - 40))))


def prompt_input(term, key, **kwargs):
    """ Prompt for user input. """
    sep_ok = getattr(term, color_primary)(u'::')
    colors = {'highlight': getattr(term, color_primary)}
    echo(u'{sep} {key:>18}: '.format(sep=sep_ok, key=key))
    entry = LineEditor(colors=colors, **kwargs).read()
    if entry is None:
        log.debug('New User Account canceled at prompt key={0}.'.format(key))
    return entry


def prompt_yesno(question):
    """ yes/no user prompt. """
    term = getterminal()
    sep = getattr(term, color_secondary)(u'**')
    colors = {'highlight': getattr(term, color_secondary)}
    echo(fixate_next(term, newlines=1))
    while True:
        echo(u'{sep} {question} [yn] ?\b\b'.format(sep=sep, question=question))
        yn = LineEditor(colors=colors, width=1).read() or u''
        if yn.lower() in (u'y', u'n'):
            return yn.lower() == u'y'
        echo(term.move_x(0) + term.clear_eol)
        echo(fixate_next(term, newlines=0))


def show_validation_error(errmsg):
    """ Display validation error message. """
    term = getterminal()
    sep_bad = getattr(term, color_secondary)(u'**')
    echo(fixate_next(term, newlines=1))
    for txt in term.wrap(errmsg, width=max(0, min(80, term.width) - 3)):
        echo(fixate_next(term, newlines=0))
        echo(u'{sep} {txt}'.format(sep=sep_bad, txt=txt))
        echo(u'\r\n')


def do_nua(user):
    """ Perform new user account field setting and validation. """
    session, term = getsession(), getterminal()
    session.activity = u'Applying for an account'

    idx = 0
    validation_fields = get_validation_fields(user)
    while idx != len(validation_fields):
        field = validation_fields.values()[idx]
        echo(fixate_next(term, newlines=1))
        if field.description:
            show_description(term, field.description,
                             color=getattr(term, color_secondary))
            echo(u'\r\n')
            echo(fixate_next(term, newlines=0))
        value = prompt_input(term=term,
                             key=field.prompt_key,
                             content=field.getter(),
                             **field.kwargs)

        # user pressed escape, prompt for cancellation
        if value is None:
            if prompt_yesno('Cancel'):
                return None
            # re-prompt for current field
            continue

        value = value.strip()

        # validate using function, `field.validation_function()'
        if field.validation_function:
            errmsg, mod = field.validation_function(user, value)
            if errmsg:
                show_validation_error(errmsg)
                # re-prompt for current field
                idx += mod
                continue

        # set local variable, increment index to next field
        if field.name:
            setattr(user, field.name, value)
        idx += 1

    return user


def main(handle=u''):
    """
    Main procedure.
    """

    # set syncterm font, if any
    term = getterminal()
    if term.kind == 'ansi':
        echo(syncterm_setfont(syncterm_font))

    # reset handle to an empty string if it is any
    # of the 'new' user account alias strings
    if handle.lower() in new_usernames:
        handle = u''

    user = User(handle)

    # create new user record for manipulation
    while True:
        display_banner(art_file, encoding=art_encoding)
        user = do_nua(user)

        # user canceled.
        if user is None:
            return

        # confirm
        if prompt_yesno(question='Create account'):
            assert not find_user(user.handle), (
                # prevent race condition, scenario: `billy' begins new account
                # process, waits at 'Create account [yn] ?' prompt until a
                # second `billy' finishes account creation process, then the
                # first `billy' enters 'y', overwriting the existing account.
                'Security race condition: account already exists')
            user.save()
            goto(top_script, user.handle)



""" Last Callers script for x/84. """
# std
import os
import time
import collections

# local
from x84.bbs import getsession, getterminal, get_ini, echo
from x84.bbs import DBProxy, timeago, syncterm_setfont
from common import prompt_pager, display_banner

#: filepath to folder containing this script
here = os.path.dirname(__file__)

#: filepath to artfile displayed for this script
art_file = os.path.join(here, 'art', 'callers*.ans')

#: encoding used to display artfile
art_encoding = 'cp437'

#: fontset for SyncTerm emulator
syncterm_font = 'topaz'

#: maximum length of user handles
username_max_length = get_ini(section='nua',
                              key='max_user',
                              getter='getint'
                              ) or 10

#: maximum length of user handles
location_max_length = get_ini(section='nua',
                              key='max_location',
                              getter='getint'
                              ) or 15

numcalls_max_length = len('# calls')

call_record = collections.namedtuple(
    'lc', ['timeago', 'num_calls', 'location', 'handle'])


def get_lastcallers(last):
    timenow = time.time()
    return sorted([call_record(timeago=timenow - time_called,
                               num_calls=num_calls,
                               location=location,
                               handle=handle.decode('utf8'))
                   for handle, (time_called, num_calls, location)
                   in DBProxy('lastcalls').items()])[:last]


def main(last=10):
    """
    Script entry point.

    :param int last: Number of last callers to display
    """
    session, term = getsession(), getterminal()
    session.activity = u'Viewing last callers'

    colors = [term.green, term.bright_blue, term.bold,
              term.cyan, term.bold_black]

    # set syncterm font, if any
    if syncterm_font and term.kind.startswith('ansi'):
        echo(syncterm_setfont(syncterm_font))

    # display banner
    line_no = display_banner(filepattern=art_file, encoding=art_encoding)

    # get last callers
    last_callers = get_lastcallers(last=last)
    echo(u'\r\n\r\n')

    def make_header(fmt):
        return fmt.format(
            handle=term.bold_underline(
                term.ljust('handle', username_max_length + 1)),
            location=term.underline(
                term.ljust('location', location_max_length)),
            num_calls=term.bold_underline(
                term.ljust('# calls', numcalls_max_length)),
            timeago=term.underline('time ago'))

    call_fmt = '{handle} {location} {num_calls} {timeago}'
    header = make_header(call_fmt)
    header_length = term.length(header)
    if header_length > term.width:
        call_fmt = '{handle} {num_calls} {timeago}'
        header = make_header(call_fmt)
        header_length = term.length(header)

    # format callers, header:
    callers_txt = [header] + [' ' * header_length]

    # content:
    callers_txt.extend([
        call_fmt.format(
            handle=lc.handle.ljust(username_max_length + 1),
            location=term.ljust(colors[idx % len(colors)](
                lc.location or '-' * location_max_length),
                location_max_length),
            num_calls='{0}'.format(
                lc.num_calls).rjust(numcalls_max_length),
            timeago=colors[idx % len(colors)](
                timeago(lc.timeago))
        ) for idx, lc in enumerate(last_callers)
    ])

    # display file contents, decoded, using a command-prompt pager.
    prompt_pager(content=callers_txt,
                 line_no=line_no + 2,
                 colors={'highlight': term.bright_green,
                         'lowlight': term.cyan, },
                 width=max(term.length(txt) for txt in callers_txt),
                 breaker=None)



""" Custom exceptions for x/84. """

# local imports
from x84.bbs.script_def import Script


class Disconnected(Exception):

    """ Thrown when a client is disconnected. """

    pass


class Goto(Exception):

    """ Thrown to change script without returning. """

    def __init__(self, script, *args, **kwargs):
        self.value = Script(name=script, args=args, kwargs=kwargs)



"""
rlogin server for x84.

This only exists to demonstrate alternative client protocols rather than
only ssh or telnet.  rlogin is a very insecure and not recommended!
"""
# http://www.ietf.org/rfc/rfc1282.txt

import logging
import select
import socket
import array
import errno
import time

# local
from x84.bbs.userbase import (
    check_new_user,
    check_bye_user,
    check_anonymous_user
)
from x84.bbs.exception import Disconnected
from x84.client import BaseClient, BaseConnect
from x84.server import BaseServer
from x84.terminal import spawn_client_session


class RLoginClient(BaseClient):

    """ rlogin protocol client handler. """

    kind = 'rlogin'

    def __init__(self, sock, address_pair, on_naws=None):
        super(RLoginClient, self).__init__(sock, address_pair, on_naws)

        # Urgent send buffer (MSG_OOB)
        self.usend_buffer = array.array('c')

    def recv_ready(self):
        """ Whether data is awaiting on the telnet socket. """
        return (self.is_active() and bool(
            select.select([self.sock.fileno()], [], [], 0)[0]))

    def send(self):
        """
        Send any data buffered and return number of bytes send.

        :raises Disconnected: client has disconnected (cannot write to socket).
        """
        if len(self.usend_buffer) > 0:
            ready_bytes = bytes(''.join(self.usend_buffer))
            self.usend_buffer = array.array('c')

            def _send_urgent(send_bytes):
                """ Sent urgent (out of band) TCP packet. """
                try:
                    return self.sock.send(send_bytes, socket.MSG_OOB)
                except socket.error as err:
                    if err.errno in (errno.EDEADLK, errno.EAGAIN):
                        self.log.debug('{self.addrport}: {err} '
                                       '(bandwidth exceed)'
                                       .format(self=self, err=err))
                        return 0
                    raise Disconnected('send: {0}'.format(err))

            sent = _send_urgent(ready_bytes)
            if sent < len(ready_bytes):
                self.usend_buffer.fromstring(ready_bytes[sent:])

        else:
            super(RLoginClient, self).send()

    def send_ready(self):
        """ Whether any data is buffered for delivery. """
        return bool(len(self.send_buffer) + len(self.usend_buffer))

    def send_urgent_str(self, bstr):
        """ Buffer urgent (OOB) message to client from bytestring. """
        self.usend_buffer.fromstring(bstr)


class ConnectRLogin(BaseConnect):

    """
    rlogin protocol connection handler.

    Takes care of the (initial) handshake, terminal and session setup.
    """

    #: maximum time elapsed allowed for on-connect negotiation
    TIME_NEGOTIATE = 5.0

    #: poll interval for on-connect negotiation
    TIME_POLL = 0.10

    def run(self):
        """
        Perform rfc1282 (rlogin) connection establishment.

        Determine rlogin on-connect data, rlogin may only
        negotiate session user name and terminal type.
        """
        try:
            self._set_socket_opts()

            self.banner()

            # Receive on-connect data-value pairs, may raise ValueError.
            data = self.get_connect_data()

            # parse into dict,
            parsed = self.parse_connect_data(data)
            for key, value in parsed.items():
                if value:
                    self.log.debug('{client.addrport}: {key}={value}'
                                   .format(client=self.client,
                                           key=key, value=value))

            # and apply to session-local self.client.env.
            self.apply_environment(parsed)

            # The server returns a zero byte to indicate that it has received
            # these strings and is now in data transfer mode.
            if self.client.is_active():
                self.client.send_str(bytes('\x00'))

                # The remote server indicates to the client that it can accept
                # window size change information by requesting a window size
                # message (as out of band data) just after connection
                # establishment and user identification exchange.  The client
                # should reply to this request with the current window size.
                #
                # Disabled: neither SyncTERM or BSD rlogin honors this, and
                # we haven't got any code to parse it. Its in the RFC but ..
                self.client.send_urgent_str(bytes('\x80'))

            matrix_kwargs = {}
            username = parsed.get('server-user-name', 'new')
            if check_new_user(username):
                # new@ login may be allowed
                matrix_kwargs['new'] = True
            if check_bye_user(username):
                # rlogin as 'bye', 'logoff', etc. not allowed
                raise ValueError('Bye user {0!r} used by rlogin'
                                 .format(username))
            if check_anonymous_user(username):
                # anonymous@ login may be allowed
                matrix_kwargs['anonymous'] = True

            if self.client.is_active():
                return spawn_client_session(client=self.client,
                                            matrix_kwargs=matrix_kwargs)
        except socket.error as err:
            self.log.debug('{client.addrport}: connection closed: {err}'
                           .format(client=self.client, err=err))
        except EOFError:
            self.log.debug('{client.addrport}: EOF from client'
                           .format(client=self.client))
        except Exception as err:
            self.log.debug('{client.addrport}: connection closed: {err}'
                           .format(client=self.client, err=err))
        finally:
            self.stopped = True
        self.client.deactivate()

    def get_connect_data(self):
        """
        Receive four null-terminated strings transmitted by client on-connect.

        :return: bytes received, containing at least 4 NUL-terminated strings.
        :rtype: str
        :raises ValueError: on-connect data timeout or bandwidth exceeded.
        """
        established_msg = ('{client.addrport}: rlogin connection established'
                           .format(client=self.client))
        data = array.array('c')

        #: maximum size of negotiation string
        MAXLEN = 4096

        err = None
        st_time = time.time()
        while True:
            # allow time to pass for more data,
            time.sleep(self.TIME_POLL)

            if self.client.recv_ready():
                self.client.socket_recv()

            if self.client.input_ready():
                # data to be received,
                # read in data.
                data.fromstring(self.client.get_input())

            n_nul = data.count('\x00')
            if n_nul >= 3:
                self.client.env['RLOGIN_CLIENT_NAME'] = {
                    3: 'SyncTERM',
                    4: 'BSD',
                }.get(n_nul, 'unknown:{0})'.format(n_nul))
                if self.client.env['RLOGIN_CLIENT_NAME'] == 'SyncTERM':
                    self.client.env['encoding'] = 'cp437'

                self.log.debug('{msg} ({env[RLOGIN_CLIENT_NAME]})'
                               .format(msg=established_msg,
                                       env=self.client.env))
                return data.tostring()

            elif time.time() - st_time >= self.TIME_NEGOTIATE:
                # too much time has elapsed, give up.
                err = 'rlogin on-connect timeout'

            elif len(data) >= MAXLEN:
                # client has sent an abusive number of bytes, disconnect.
                err = 'rlogin bandwidth exceeded'

            if err:
                raise ValueError(err)

    def apply_environment(self, parsed):
        """
        Cherry-pick rlogin values into client environment variables.

        :param dict parsed: values identified by class method
                            ``parse_connect_data()``
        :rtype: None
        """
        # Only terminal-type environment variable is propagated from client, we
        # can also discern their USER by 'client-user-name', which we would
        # expect to be analogous to telnet environment value USER.
        self.client.env['TERM'] = parsed.get('terminal-type', 'vt220')
        if 'client-user-name' in parsed:
            self.client.env['USER'] = parsed.get('client-user-name')

    def parse_connect_data(self, data):
        """
        Parse and return raw data received by client on-connect.

        :param str data: bytes received by class method get_connect_data().
        :return: dictionary containing pertinent key/values
        :rtype: dict
        """
        parsed = dict()
        try:
            # Upon connection establishment, the client sends four
            # null-terminated strings to the server.  The first is an empty
            # string (i.e., it consists solely of a single zero byte), followed
            # by three non-null strings: the client username, the server
            # username, and the terminal type and speed.  More explicitly:
            #
            #        <null>
            #        client-user-name<null>
            #        server-user-name<null>
            #        terminal-type/speed<null>
            segs = data.split('\x00')

            # SyncTerm leaves a null-terminating byte.
            if len(segs) == 5 and segs[4] == '':
                segs.pop(4)

            for segname in ('null',
                            'client-user-name',
                            'server-user-name',
                            'terminal-type/speed'):
                if len(segs):
                    parsed[segname] = segs.pop(0)

            parsed['terminal-type'], parsed['terminal-speed'] = (
                parsed.pop('terminal-type/speed', 'unknown/0')
                .split('/', 2))
        except ValueError as err:
            self.log.exception("ValueError in parse_connect_data: {err}"
                               .format(err=err))

        return parsed

    def _set_sock_opts(self):
        """ Set the socket in non-blocking mode. """
        self.client.sock.setblocking(0)
        self.client.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)


class RLoginServer(BaseServer):

    """ RLogin/RSH protocol server. """

    client_factory = RLoginClient
    connect_factory = ConnectRLogin

    def __init__(self, config):
        """ Class initializer. """
        self.log = logging.getLogger(__name__)
        self.config = config
        self.addr = config.get('rlogin', 'addr')
        self.port = 513
        if config.has_option('rlogin', 'port'):
            # rlogin is coded for port 513, though you could specify an
            # alternative port if you really wished.
            self.port = config.getint('rlogin', 'port')

        # bind
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET,
                                      socket.SO_REUSEADDR, 1)
        try:
            self.server_socket.bind((self.addr, self.port))
            self.server_socket.listen(self.LISTEN_BACKLOG)
        except socket.error as err:
            self.log.error('unable to bind {self.addr}:{self.port}: {err}'
                           .format(self=self, err=err))
            exit(1)

        self.log.info('rlogin listening on {self.addr}:{self.port}/tcp'
                      .format(self=self))

    def client_fds(self):
        """ Return list of rlogin client file descriptors. """
        fds = [client.fileno() for client in self.clients.values()]
        # pylint: disable=bad-builtin
        #         You're drunk, pylint
        return filter(None, fds)



# -*- coding: utf-8 -*-
""" Password reset script for x/84. """
from __future__ import division

import string
import random

from x84.bbs import getsession, getterminal
from x84.bbs import echo, LineEditor
from x84.bbs import get_ini, find_user, get_user
from common import display_banner

import logging
import smtplib
import base64
import os
from email.mime.text import MIMEText

log = logging.getLogger(__name__)

here = os.path.dirname(__file__)

#: maximum length of user handles
username_max_length = get_ini(section='nua',
                              key='max_user',
                              getter='getint'
                              ) or 10

#: maximum length of password
password_max_length = get_ini(section='nua',
                              key='max_pass',
                              getter='getint'
                              ) or 15

#: maximum length of email address
email_max_length = get_ini(section='nua',
                           key='max_email',
                           getter='getint'
                           ) or 40

#: smtp host for mail delivery
mail_smtphost = get_ini(section='system', key='mail_smtphost'
                        ) or 'localhost'

#: body of password reset message
msg_mailbody = (u'A password reset has been requested on {bbsname} '
                u'from {session.sid!r} by matching your user handle '
                u'{user.handle!r} and e-mail address.\r\n\r\n'
                u'Your password reset key is {passkey!r}')

msg_mailsubj = u'Passkey token for {bbsname}'

msg_mailfrom = get_ini(section='system',
                       key='mail_addr'
                       ) or 'sysop@localhost.localdomain'

#: name of bbs
system_bbsname = get_ini(section='system',
                         key='bbsname'
                         ) or 'Unnamed'

#: banner art file
art_file = os.path.join(here, 'art', 'pwreset.asc')

#: primary color (highlight)
color_primary = 'red'

#: secondary color (lowlight)
color_secondary = 'bold_black'

#: maximum attempts for passkey authentication
passkey_max_attempts = 5

#: password hidden character
hidden_char = u'\u00f7'

#: positions next prompt 40 pixels minus center of screen
fixate_next = lambda term: (
    u'\r\n\r\n' + term.move_x(max(0, (term.width // 2) - 40)))


def display_banner_animation(banner_text):
    """
    Display animation of "cracking" the banner text.

    This is inspired by all those movies.
    """
    session, term = getsession(), getterminal()
    highlight = getattr(term, color_primary)
    lowlight = getattr(term, color_secondary)
    animation_length = len(banner_text)
    animation_speed = 0.1

    def get_garbage():
        # get some random "cracking" strings
        txt = list(string.printable)[:len(banner_text)]
        random.shuffle(txt)
        return txt

    get_x = lambda: (
        # get x-position of banner (center)
        (term.width // 2) - (len(banner_text) // 2))

    def display_header(xpos, banner_text):
        x_top_left = xpos - 1
        x_bot_right = xpos + (len(banner_text) - 5)
        return u'\r\n'.join((
            u'{xpos}{txt}'.format(xpos=term.move_x(x_top_left),
                                  txt=lowlight(u'┬─────')),
            u'',
            u'{xpos}{txt}'.format(xpos=term.move_x(x_bot_right),
                                  txt=lowlight(u'─────┴')),
        ))

    def decorate_guess(guess, actual):
        # return string where matching letters are highlighted
        attr = None
        rstr = u''
        for idx, ch_guess in enumerate(guess):
            # optimized attribute draws
            if ch_guess == actual[idx]:
                if attr != highlight:
                    attr = highlight
                    rstr += term.normal + attr
            else:
                if attr != lowlight:
                    attr = lowlight
                    rstr += term.normal + attr
            rstr += ch_guess
        return rstr

    def merge_garbage(prior_guess, garbage, actual):
        # return string with new garbage mixed in,
        # except where already matching
        next_guess = garbage[:]
        for idx, garbage_item in enumerate(garbage):
            if prior_guess[idx] == actual[idx]:
                next_guess[idx] = actual[idx]
            else:
                next_guess[idx] = garbage_item
        return next_guess

    def make_match(guess, actual):
        next_guess = guess[:]
        indicies = range(len(actual))
        random.shuffle(indicies)
        for idx in indicies:
            if next_guess[idx] != actual[idx]:
                next_guess[idx] = actual[idx]
                break
        return next_guess

    xpos = get_x()

    # display header,
    echo(display_header(xpos, banner_text))

    # move-to banner animation row
    echo(term.move_up())

    guess = get_garbage()
    for _ in range(0, animation_length):
        # check for screen resize
        if session.poll_event('refresh'):
            # remove artifacts, get new center
            echo(term.move_x(0) + term.clear_eol)
            xpos = get_x()

        echo(term.move_x(xpos))
        echo(decorate_guess(guess=guess, actual=banner_text))
        echo(term.clear_eol)

        if guess == banner_text:
            # "cracked"
            break

        if term.inkey(timeout=animation_speed):
            # user canceled animation
            break

        # mix in new garbage
        guess = merge_garbage(prior_guess=guess,
                              garbage=get_garbage(),
                              actual=banner_text)

        # ensure at least one new index is guessed
        guess = make_match(guess=guess,
                           actual=banner_text)

    # end of animation
    echo(term.move_x(xpos))
    echo(highlight(banner_text))
    echo(term.move_x(0) + (term.move_down * 2))


def prompt_input(term, key, **kwargs):
    """ Prompt for user input. """
    sep_ok = getattr(term, color_secondary)(u'::')
    sep_bad = getattr(term, color_primary)(u'::')
    colors = {'highlight': getattr(term, color_primary)}

    echo(fixate_next(term))
    echo(u'{sep} {key:>18}: '.format(sep=sep_ok, key=key))
    entry = LineEditor(colors=colors, **kwargs).read() or u''
    if not entry.strip():
        echo(fixate_next(term))
        echo(u'{sep} Canceled !\r\n'.format(sep=sep_bad))
        log.debug('Password reset canceled at prompt key={0}.'.format(key))
        return u''

    return entry


def matches_email(handle, email):
    """ Return User record for handle if it matches email. """
    matching_handle = find_user(handle)
    user = matching_handle and get_user(matching_handle)
    if not user:
        log.debug('password reset failed, no such user {0} for email {1}.'
                  .format(handle, email))
        return False
    elif not user.email.strip():
        log.debug('password reset failed, user {0} has no email on file.'
                  .format(handle))
        return False
    elif email.lower() != user.email.lower():
        log.debug('pasword reset failed, email mismatch: {0} != {1}.'
                  .format(email, user.email))
        return False

    # success !
    return user


def send_passkey(user):
    """ Send passkey token to user by e-mail. """
    session = getsession()
    passkey = base64.encodestring(os.urandom(50))[:password_max_length]

    email_msg = MIMEText(msg_mailbody.format(bbsname=system_bbsname,
                                             session=session,
                                             user=user,
                                             passkey=passkey))
    email_msg['From'] = msg_mailfrom
    email_msg['To'] = user.email
    email_msg['Subject'] = msg_mailsubj.format(bbsname=system_bbsname)

    try:
        smtp = smtplib.SMTP(mail_smtphost)
        smtp.sendmail(msg_mailfrom, [user.email], email_msg.as_string())
        smtp.quit()
    except Exception as err:
        log.exception(err)
        echo(u'{0}'.format(err))
        return False

    log.info(u'Password reset token delivered '
             u'to address {0!r} for user {1!r}.'
             .format(user.email, user.handle))
    return passkey


def do_reset(term, handle, email=u''):
    """ Password reset by e-mail loop. """
    sep_ok = getattr(term, color_secondary)(u'::')
    sep_bad = getattr(term, color_primary)(u'::')
    email = u''

    for _ in range(passkey_max_attempts):
        handle = prompt_input(term=term,
                              key='Username',
                              content=handle or u'',
                              width=username_max_length)

        if not handle:
            # canceled
            return False

        email = prompt_input(term=term,
                             key='E-mail',
                             content=email or u'',
                             width=email_max_length)
        if not email:
            # canceled
            return False

        user = matches_email(handle, email)
        if not user:
            echo(fixate_next(term))
            echo(u'{0} Address is incorrect !'.format(sep_bad))
            # try e-mail address again
            continue

        echo(fixate_next(term))
        passkey = send_passkey(user)
        if not passkey:
            # failed to send e-mail
            term.inkey(1)
            echo(u'\r\n\r\n')
            return False

        echo(u'{0} E-mail successfully delivered !'.format(sep_ok))

        for _ in range(passkey_max_attempts):
            try_passkey = prompt_input(term=term,
                                       key='Passkey',
                                       width=password_max_length)

            if not try_passkey:
                # canceled
                return False

            if passkey.strip() != try_passkey.strip():
                # passkey does not match
                echo(fixate_next(term))
                echo(u'{0} Passkey does not verify !'.format(sep_bad))
                # try passkey again
                continue

            new_password = prompt_input(term=term,
                                        key='Password',
                                        hidden=hidden_char,
                                        width=password_max_length)
            if not new_password:
                # canceled
                return False

            user.password = new_password
            user.save()
            log.debug('password reset successful for user {0!r}.'
                      .format(user.handle))
            echo(fixate_next(term))
            echo(u'{0} Password reset successful !'.format(sep_ok))
            return True

        echo(fixate_next(term))
        echo(u'{0} Too many authentication attempts.'.format(sep_bad))

    echo(fixate_next(term))
    echo(u'{0} Too many authentication attempts.'.format(sep_bad))


def main(handle=None):
    """
    Main procedure.
    """
    session, term = getsession(), getterminal()
    session.activity = u'resetting password'

    display_banner(art_file)

    # mix animation into art
    echo(term.move_up() * 5)

    # display banner animation
    display_banner_animation(banner_text=u'reset account password')

    return do_reset(term, handle)



""" Tetris game for x/84. """
__author__ = 'Johannes Lundberg'
__copyright__ = 'Copyright 2007-2013'
__license__ = 'Public Domain'
# Single player tetris, originally written for The Progressive (prsv), the
# predecessor python bbs system of x/84.
# Copyright (C) 2007-2013 Johannes Lundberg


def main():
    from x84.bbs import getsession, getterminal, echo
    session, term = getsession(), getterminal()
    session.activity = 'playing tetris'

    if term.width < 79 or term.height < 24:
        echo(u'Sorry, your terminal is not large enough (72x22), ')
        echo(u'your size: {term.width}x{term.height}'.format(term=term))
        echo(u'\r\n\r\nPress any key.')
        term.inkey()
        return

    with term.hidden_cursor():
        score = play()
        if score[0] > 0:
            register_score(session.user.handle, score)
        show_scores()


def register_score(handle, score):
    from x84.bbs import DBProxy
    db = DBProxy('tetris')
    if not handle in db:
        db[handle] = score
    elif score[0] > db[handle][0]:
        db[handle] = score


def show_scores():
    from x84.bbs import DBProxy, Pager, getterminal
    from x84.bbs import echo, getsession, ini
    session, term = getsession(), getterminal()
    allscores = DBProxy('tetris').items()
    if 0 == len(allscores):
        return
    # line up over tetris game, but logo & 'made by jojo' in view
    # -- since we have so much screen width, columize the scores,
    # the math brings it out to 2 columns, but fmt is adjustable
    pager_title = term.blue_reverse_underline('- hiGh SCOREs -')
    len_handle = ini.CFG.getint('nua', 'max_user')
    score_fmt = u'%s %s %s %s'
    len_pos = 2
    len_score = 10
    len_level = 3
    height, width = 11, 73
    yloc, xloc = 10, 3
    pager = Pager(height=height, width=width, yloc=yloc, xloc=xloc)
    pager.xpadding = 1
    pager.glyphs['left-vert'] = u''
    pager.glyphs['right-vert'] = u''
    pager.colors['border'] = term.blue_reverse
    pager.alignment = 'center'
    # pre-fesh pager border before fetch
    echo(pager.border() + pager.title(pager_title) + pager.clear())
    highscores = sorted(
        [(_score, _level, _handle.decode('utf8'))
         for (_handle, (_score, _level, _)) in allscores],
        reverse=True)
    pager.append(score_fmt % (
        term.bold_blue_underline('No'.ljust(len_pos)),
        term.bold_blue_underline('SCORE'.ljust(len_score)),
        term.bold_blue_underline('lVl'.ljust(len_level)),
        term.bold_blue_underline('hANdlE'.rjust(len_handle),)))
    pos = 0
    for pos, (_score, _level, _handle) in enumerate(highscores):
        if _handle == session.user.handle:
            pager.append(score_fmt % (
                term.blue_reverse(str(pos + 1).ljust(len_pos)),
                term.blue_reverse(str(_score).ljust(len_score)),
                term.blue_reverse(str(_level).ljust(len_level)),
                term.blue_reverse(_handle.rjust(len_handle)),))
        else:
            pager.append(score_fmt % (
                term.bold_blue(str(pos + 1).ljust(len_pos)),
                term.blue(str(_score).ljust(len_score)),
                term.blue(str(_level).ljust(len_level)),
                _handle.rjust(len_handle),))
    # append additional empty slot rows
    while len(pager.content) < pager.visible_height:
        pos += 1
        pager.append(score_fmt % (
            term.bold_blue(str(pos + 1).ljust(len_pos)),
            term.bold_black(u'.'.ljust(len_score)),
            term.bold_black(u'.'.ljust(len_level)),
            term.bold_black(u'.'.rjust(len_handle)),))

    dirty = 2  # 2=do not refresh border & title
    pager.move_home()
    while not pager.quit:
        # 1=full refresh
        dirty = 1 if session.poll_event('refresh') else dirty
        if dirty:
            echo(u''.join((
                term.normal,
                ((pager.border() + pager.title(pager_title))
                    if dirty != 2 else u''),
                pager.refresh(),
                pager.footer(u''.join((
                    term.underline_blue('q'),
                    term.bold_blue('uit')))),
            )))
            dirty = 0
        echo(pager.process_keystroke(term.inkey(1)))


def play():
    import time
    from random import randint
    import os
    from x84.bbs import getterminal, from_cp437, AnsiWindow, syncterm_setfont
    from x84.bbs import echo as echo_unbuffered
    term = getterminal()
    field = []
    global charcache
    charcache = u''
    field_width = 10
    field_height = 20
    # Access scheme looks like this:
    #   layout[p][r][ypox][xpos]
    # layoutcolor = [ 7,2,3,4,4,6,7 ]
    layout = [
        #  ##
        #  ##
        [
            [
                [1, 1, ],
                [1, 1, ],
            ],
        ],
        #  #
        #  #
        #  #
        #  #
        [
            [
                [0, 1, 0, 0],
                [0, 1, 0, 0],
                [0, 1, 0, 0],
                [0, 1, 0, 0],
            ],
            [
                [0, 0, 0, 0],
                [1, 1, 1, 1],
                [0, 0, 0, 0],
                [0, 0, 0, 0],
            ]
        ],
        #  ###
        #   #
        [
            [
                [0, 0, 0],
                [1, 1, 1],
                [0, 1, 0],
            ],
            [
                [0, 1, 0],
                [0, 1, 1],
                [0, 1, 0],
            ],
            [
                [0, 1, 0],
                [1, 1, 1],
                [0, 0, 0],
            ],
            [
                [0, 1, 0],
                [1, 1, 0],
                [0, 1, 0],
            ],
        ],
        #  #
        #  #
        #  ##
        [
            [
                [0, 1, 0],
                [0, 1, 0],
                [0, 1, 1],
            ],
            [
                [0, 0, 1],
                [1, 1, 1],
                [0, 0, 0],
            ],
            [
                [1, 1, 0],
                [0, 1, 0],
                [0, 1, 0],
            ],
            [
                [0, 0, 0],
                [1, 1, 1],
                [1, 0, 0],
            ],
        ],
        #   #
        #   #
        #  ##
        [
            [
                [0, 1, 0],
                [0, 1, 0],
                [1, 1, 0],
            ],
            [
                [0, 0, 0],
                [1, 1, 1],
                [0, 0, 1],
            ],
            [
                [0, 1, 1],
                [0, 1, 0],
                [0, 1, 0],
            ],
            [
                [1, 0, 0],
                [1, 1, 1],
                [0, 0, 0],
            ],
        ],
        #  ##
        #   ##
        [
            [
                [0, 1, 0],
                [1, 1, 0],
                [1, 0, 0],
            ],
            [
                [0, 0, 0],
                [1, 1, 0],
                [0, 1, 1],
            ],
        ],
        #   ##
        #  ##
        [
            [
                [0, 1, 0],
                [0, 1, 1],
                [0, 0, 1],
            ],
            [
                [0, 0, 0],
                [0, 1, 1],
                [1, 1, 0],
            ],
        ],
    ]

    fieldx1 = 32
    fieldy1 = 10
    scorex1 = 11
    scorey1 = 11

    class RectRedraw:
        x1 = None
        y1 = None
        x2 = None
        y2 = None

        def max(r, val, valmax):
            if val > valmax:
                return valmax
            return val

        def min(r, val, valmin):
            if val < valmin:
                return valmin
            return val

        def merge(r, x1, y1, x2, y2):
            if r.x1 is None or r.x1 > x1:
                r.x1 = r.min(x1, 0)
            if r.y1 is None or r.y1 > y1:
                r.y1 = r.min(y1, 0)
            if r.x2 is None or r.x2 < x2:
                r.x2 = r.max(x2, field_width)
            if r.y2 is None or r.y2 < y2:
                r.y2 = r.max(y2, field_height)
            # print r.x1,r.y1,r.x2,r.y2

        def clean(r):
            r.x1 = None
            r.y1 = None
            r.x2 = None
            r.y2 = None
    rr = RectRedraw()
    for _ in range(field_height):
        field.append([0] * field_width)

    def echo(s):
        global charcache
        charcache += s
    assert term.height > (field_height + 1)
    echo_unbuffered(u''.join((
        u'\r\n\r\n',
        u'REAdY YOUR tERMiNAl %s ' % (term.bold_blue('(!)'),),
        u'\r\n\r\n',
        u'%s PRESS ANY kEY' % (term.bold_black('...'),),
    )))

    term.inkey()

    # set syncterm font to cp437
    if term.kind.startswith('ansi'):
        echo_unbuffered(syncterm_setfont('cp437'))
    artfile = os.path.join(os.path.dirname(__file__), 'art', 'tetris.ans')
    echo_unbuffered(u'\r\n' * term.height)  # cls
    if os.path.exists(artfile):
        echo_unbuffered(from_cp437(open(artfile).read()).rstrip())

    def gotoxy(x, y):
        echo(term.move(y, x))

    def plotblock(color, lastcolor):
        if color:
            c = u'\u2588\u2588'  # '\xDB\xDB'
        else:  # both empty
            c = '  '
            color = 0
        # Output optimization
        if color % 8 == 0:
            color = color / 8
        if color == lastcolor:
            echo(c)
        else:
            if color:
                fg = str(30 + color % 8)
            else:
                fg = '37'
            if color >= 8:
                bg = ';%d' % (40 + color / 8)
            else:
                bg = ''
            echo('\x1b[0;' + fg + bg + 'm')
            echo(c)
            lastcolor = color
        return lastcolor

    def drawfield():
        lastcolor = ''
        for y in range(0, field_height, 2):
            # gotoxy(field_width,2+y/2)
            gotoxy(fieldx1 + 2, fieldy1 + 1 + y / 2)
            # Which block to show, full, half-up, half-down or empty.
            for x in range(field_width):
                color = field[y][x] + field[y + 1][x] * 8
                if field[y][x] and field[y + 1][x]:
                    c = u'\u2588'  # '\xDB'
                    if field[y][x] == field[y + 1][x]:
                        color = color % 8
                    else:
                        c = u'\u2580'  # '\xDF'
                elif field[y][x] and not field[y + 1][x]:
                    c = u'\u2580'  # '\xDF'
                elif not field[y][x] and field[y + 1][x]:
                    c = u'\u2584'  # '\xDC'
                else:  # both empty
                    c = ' '
                # Output optimization
                if color % 8 == 0:
                    color = color / 8
                if color == lastcolor:
                    echo(c)
                else:
                    if color:
                        fg = str(30 + color % 8)
                    else:
                        fg = '37'
                    if color >= 8:
                        bg = ';%d' % (40 + color / 8)
                    else:
                        bg = ''
                    echo('\x1b[0;' + fg + bg + 'm')
                    echo(c)
                    lastcolor = color
        echo(term.normal)

    layoutcolor = [7, 2, 7, 6, 3, 6, 3]
    # p    = -1  # Current piece type
    nextpiece = randint(0, len(layout) - 1)
    p = randint(0, len(layout) - 1)
    p = 1
    r = 0   # Current rotation
    xpos = 4   # X position
    # ypos = -2  # Y position
    ypos = -len(layout[p][0])
    level = 1
    score = 0
    lines = 0

    def flush():
        global charcache
        echo_unbuffered(charcache)
        charcache = u''

    def fillpiece(x, y, p, r, value):
        row = 0
        for line in layout[p][r]:
            col = 0
            for c in line:
                if c and (y + row) >= 0:
                    field[y + row][x + col] = value
                col += 1
            row += 1

    def showpiece(x, y, p, r):
        fillpiece(x, y, p, r, layoutcolor[p])

    def hidepiece():
        fillpiece(xpos, ypos, p, r, 0)

    def testpiece(x, y, newr):
        hidepiece()
        # Space at the new location?
        row = 0
        for line in layout[p][newr]:
            col = 0
            for c in line:
                try:
                    if c:
                        if ((y + row) >= 0 and field[y + row][x + col]
                                or (x + col) < 0 or (x + col) > 9):
                            return 0
                except IndexError:
                    return 0
                col += 1
            row += 1
        # Movement possible
        return 1

    def movepiece(x, y, newr):
        if testpiece(x, y, newr):
            # Build redraw rectangle
            rr.merge(xpos, ypos,
                     xpos + len(layout[p][0][0]), ypos + len(layout[p][0]))
            rr.merge(x, y,
                     x + len(layout[p][0][0]), y + len(layout[p][0]))
            showpiece(x, y, p, newr)
            return (x, y, newr, 1)
        else:
            showpiece(xpos, ypos, p, r)
            return (xpos, ypos, r, 0)

    def shownext(p):
        r = 0
        for y in range(4):
            gotoxy(26, 18 + y)
            echo(u' ' * 4)

        echo(term.color(layoutcolor[p]))

        yoffset = int(len(layout[p][r][0]) < 4)
        xoffset = int(len(layout[p][r]) < 3)
        for y in range(len(layout[p][r])):
            for x in range(len(layout[p][r][y])):
                val = layout[p][r][y][x]
                if val:
                    gotoxy(26 + x + xoffset, 18 + y + yoffset)
                    echo(u'\u2588\u2588')

    def drawstats():
        echo(term.move(scorey1, scorex1) + '%d' % level)
        echo(term.move(scorey1 + 2, scorex1) + '%d' % lines)
        echo(term.move(scorey1 + 3, scorex1) + '%d' % score)

    drawstats()
    ticksize = 0.4
    nexttick = time.time() + ticksize
    showpiece(xpos, ypos, p, r)
    gotoxy(26, 17)
    echo(term.blue_reverse('next'))
    shownext(nextpiece)

    # Full redraw first frame
    rr.merge(0, 0, field_width, field_height)

    buf = ''
    while True:
        drawfield()
        # gotoxy(0,0)
        # echo('\x1b[37mx: %d, y: %d, p: %d         '%(xpos,ypos,p))
        slice = nexttick - time.time()
        if slice < 0:
            slice = 0
        echo(buf)
        buf = ''
        flush()
        key = term.inkey(slice + 0.01)
        now = time.time()
        # hidepiece()
        if key is not None:
            if key in (u'q', u'Q'):
                return (0, 0, 0)
            elif key.code == term.KEY_LEFT or key == u'h':
                xpos, ypos, r, m = movepiece(xpos - 1, ypos, r)
            elif key.code == term.KEY_RIGHT or key == u'l':
                xpos, ypos, r, m = movepiece(xpos + 1, ypos, r)
            elif key.code == term.KEY_UP or key == u'k':
                xpos, ypos, r, m = movepiece(
                    xpos, ypos, (r + 1) % len(layout[p]))
            elif key.code == term.KEY_DOWN or key == u'j':
                xpos, ypos, r, m = movepiece(xpos, ypos + 1, r)
            elif key in (' ',):
                m = True
                c = 0
                while m:
                    xpos, ypos, r, m = movepiece(xpos, ypos + 1, r)
                    if m:
                        c += 1
                if c:
                    nexttick = time.time() + ticksize
        # New tick?
        if now > nexttick:
            nexttick += ticksize
            # Move down piece
            xpos, ypos, r, moved = movepiece(xpos, ypos + 1, r)
            # Piece has touched down?
            if not moved:
                # Is the player dead?
                if ypos <= -len(layout[p][0]):
                    death_win = AnsiWindow(height=6, width=40,
                                           yloc=fieldy1 + 10 / 2, xloc=fieldx1 - 11)
                    death_win.colors['border'] = term.bold_black
                    echo_unbuffered(death_win.clear() + death_win.border())
                    echo_unbuffered(
                        term.move(fieldy1 + 10 / 2 + 1, fieldx1 - 11))
                    echo_unbuffered((
                                    u'!! gAME OVeR!! Score was: %i' % (score,)).center(40))
                    echo_unbuffered(
                        term.move(fieldy1 + 10 / 2 + 3, fieldx1 - 11))
                    echo_unbuffered(u'press RETURN'.center(40))
                    while True:
                        inp = term.inkey()
                        if inp.code == term.KEY_ENTER:
                            break
                    return (score, level, lines)

                # Any complete rows to remove?
                complete = []
                for y in range(field_height):
                    x = 0
                    while x < field_width:
                        if field[y][x] == 0:
                            break
                        x += 1
                    if x == field_width:
                        complete.append(y)
                if len(complete) > 0:
                    # Add score
                    lines += len(complete)
                    score += len(complete) * len(complete) * 100
                    # Shrink field
                    for line in complete:
                        del field[line]
                        field.insert(0, [0] * field_width)

                    if lines >= level * 10:
                        level += 1
                        ticksize = 0.4 - level * 0.02
                    drawstats()

                    # Redraw complete field
                    rr.merge(0, 0, field_width, field_height)

                # Time for a new piece
                p = nextpiece
                nextpiece = randint(0, len(layout) - 1)
                r = 0
                xpos = 4
                ypos = -len(layout[p][0])
                showpiece(xpos, ypos, p, r)
                shownext(nextpiece)



# coding=utf-8
"""
one-liners script for x/84.

To use the (optional) http://shroo.ms API,
which provides something of an intra-bbs oneliners,
configure a section in your .ini file::

    [shroo-ms]
    enabled = yes
    idkey = id-key-here-ask-frost-lol
    restkey = rest-key-here-ask-frost-too

You'll have to contact a guy named 'frost',
more than likely found on telnet://bbs.shroo.ms.
"""
import threading
import logging
import time

from x84.bbs import getsession, getterminal, echo, syncterm_setfont, LineEditor
from x84.bbs import timeago, decode_pipe
from x84.bbs import DBProxy, get_ini
from common import display_banner

#: maximum username length of bbs
username_max_length = get_ini(
    section='nua', key='max_user', getter='getint'
) or 10

#: filepath to artfile displayed for this script
art_file = get_ini(
    section='oneliners', key='art_file'
) or 'art/oneliner.ans'

#: encoding used to display artfile
art_encoding = get_ini(
    section='oneliners', key='art_encoding'
) or 'cp437'

#: fontset for SyncTerm emulator
syncterm_font = get_ini(
    section='oneliners', key='syncterm_font'
) or 'topaz'

#: alternating colors for oneliners text
color_palette = get_ini(
    section='oneliners', key='color_palette', split=True
) or ['bright_white', 'bright_cyan', 'bright_magenta']

#: maximum length of message posting
MAX_MSGLEN = get_ini(
    section='oneliners', key='max_msglen', getter='getint'
) or 50

#: minimum seconds elapsed between posts, 0 to disable
MIN_ELAPSED = get_ini(
    section='oneliners', key='min_elapsed', getter='getint'
) or 0

#: maximum records to display, no matter the screen size
MAX_HISTORY = get_ini(
    section='oneliners', key='max_history', getter='getint'
) or 1000

#: whether shroo.ms api is enabled
shroo_ms_enabled = get_ini(
    section='shroo-ms', key='enabled', getter='getboolean')

#: shroo.ms api url
shroo_ms_api_url = get_ini(
    section='shroo-ms', key='api_url'
) or 'https://api.parse.com/1/classes/wall'

#: shroo.ms api key
shroo_ms_api_key = get_ini(
    section='shroo-ms', key='idkey')

#: shroo.ms rest key
shroo_ms_restkey = get_ini(
    section='shroo-ms', key='restkey')

#: shroo.ms named bbs for posts
system_bbsname = get_ini(
    section='system', key='bbsname')

#: returns time of given timestamp for keysort
keysort_by_datetime = lambda oneliner: (
    time.strptime(oneliner['timestamp'], '%Y-%m-%d %H:%M:%S'))


# prevents: ImportError: Failed to import _strptime because the import lock is
# held by another thread. http://bugs.python.org/issue7980
_ = time.strptime('20110101', '%Y%m%d')  # noqa


class FetchUpdatesShrooMs(threading.Thread):

    """ Fetch shroo.ms onliners as a background thread. """

    def __init__(self):
        # It's not safe to receive data from a database from
        # within a thread ... something we should address ..
        # so we do it in __init__, bug #154
        self.existing_content = DBProxy('oneliner').copy()
        self.new_content = None
        self.log = logging.getLogger(__name__)
        super(FetchUpdatesShrooMs, self).__init__()

    def run(self):
        import requests
        params = {'limit': MAX_HISTORY, 'order': '-createdAt'}
        headers = {
            'X-Parse-Application-Id': shroo_ms_api_key,
            'X-Parse-REST-API-Key': shroo_ms_restkey,
            'Content-Type': 'application/json'
        }
        # It's ok for a thread to raise an exception naturally on request,
        result = requests.get(shroo_ms_api_url, params=params, headers=headers)
        if 200 != result.status_code:
            # log non-200 error code
            self.log.error('[shroo.ms] returned %d:', result.status_code)
            self.log.error(result.content)
            return
        content = self.transform_result(result)
        self.notify_update_database(content)

    def notify_update_database(self, content):
        new_content = {}
        for key, oneliner in content:
            if str(key) not in self.existing_content:
                new_content[key] = oneliner
        if new_content:
            self.log.info('[shroo.ms] %d new entries', len(new_content))
            self.new_content = new_content
            getsession().buffer_event('oneliner', True)
        else:
            self.log.debug('[shroo.ms] no new entries')

    def transform_result(self, result):
        # strangely, convert... base-36 integers .. haliphax's work ?
        content = [(str(int(item['objectId'], 36)), dict(
            oneliner=item['bbstagline'],
            alias=item['bbsuser'],
            bbsname=self.parse_bbsname(item['bbsname']),
            timestamp=self.parse_iso8601(item['createdAt'])
        )) for item in result.json()['results']]
        return content

    @staticmethod
    def parse_bbsname(bbsname):
        # I like the aliases, rename hellbeard's
        if bbsname == 'Blood Island/x':
            return 'BI/X'
        # and maze's
        elif bbsname == 'Random Noize':
            return 'RNz'
        # and spidy's
        elif bbsname == 'BO':
            return 'BoF'
        return bbsname

    @staticmethod
    def parse_iso8601(timestamp):
        from dateutil.parser import parse
        from dateutil.tz import tzlocal
        as_localtime = parse(timestamp).astimezone(tzlocal())
        return as_localtime.strftime('%Y-%m-%d %H:%M:%S')


def post_shroo_ms(message, username):
    """ Post to shroo.ms oneliners API. """
    # std
    import json

    # 3rd-party
    import requests

    log = logging.getLogger(__name__)

    headers = {
        'X-Parse-Application-Id': shroo_ms_api_key,
        'X-Parse-REST-API-Key': shroo_ms_restkey,
        'Content-Type': 'application/json'
    }
    payload = json.dumps({
        'bbsname': system_bbsname,
        'bbsuser': username,
        'bbstagline': message,
        # There is an ability to post as an alternative identity, presumably so
        # cap'n'hood of black flag bbs can impersonate people?  Anyway, we
        # always just set False, and we don't care on receipt whether the user
        # is impersonated.
        'bbsfakeuser': False,
    })
    try:
        result = requests.post(shroo_ms_api_url, data=payload, headers=headers)
    except Exception as err:
        # log exception string, cause message to post locally
        log.warn(err)
        return False
    else:
        if result.status_code >= 400:
            # log exceptions string, cause message to post locally
            log.error('[shroo.ms] returned %d:', result.status_code)
            log.error(result.content)
            return False
        log.info('(%d) OK [shroo.ms]: %s', result.status_code, message)

        # fetch our own post on shroo-ms in background
        thread = FetchUpdatesShrooMs()
        thread.start()
        return thread


# -- database functions


def do_merge_shroo_ms(new_content):
    """ Add oneliners from shroo-ms to local database. """
    udb = DBProxy('oneliner')
    with udb:
        udb.update(new_content)
    maybe_expunge_records()


def maybe_expunge_records():
    """ Check ceiling of database keys; trim-to MAX_HISTORY. """
    udb = DBProxy('oneliner')
    expunged = 0
    with udb:
        if len(udb) > MAX_HISTORY + 10:
            contents = DBProxy('oneliner').copy()
            _sorted = sorted(
                ((value, key) for (key, value) in contents.items()),
                key=lambda _valkey: keysort_by_datetime(_valkey[0]))
            for expunged, (_, key) in enumerate(
                    _sorted[len(udb) - MAX_HISTORY:]):
                del udb[key]
    if expunged:
        log = logging.getLogger(__name__)
        log.info('expunged %d records from database', expunged)


def add_oneline(session, message):
    """ Add a oneliner to the local database. """
    udb = DBProxy('oneliner')
    with udb:
        key = max([int(key) for key in udb.keys()] or [0]) + 1
        udb[key] = {
            'oneliner': message,
            'alias': getsession().user.handle,
            'bbsname': get_ini('system', 'bbsname'),
            'timestamp': time.strftime('%Y-%m-%d %H:%M:%S'),
        }
    maybe_expunge_records()

    # tell everybody a new oneliner was posted, including our
    # -- allows it to work something like a chatroom.
    session.send_event('global', ('oneliner', True))


# -- ui functions


def get_keymap(term, offset, height):
    """ Returns a tuple of keystroke: offset mappings """
    seq = {
        term.KEY_BACKSPACE: offset + 1,
        term.KEY_DELETE: offset + 1,
        term.KEY_UP: offset + 1,
        term.KEY_DOWN: offset + -1,
        term.KEY_ENTER: offset + -1,
        term.KEY_PGUP: offset + height - 1,
        term.KEY_PGDOWN: offset + ((height - 1) * -1),
        term.KEY_HOME: MAX_HISTORY,
        term.KEY_END: 0,
    }
    # vanilla keymap is just a mapping of ascii characters and their simulated
    # keystroke codes, which just maps back into seq{}.
    van = {
        u'k': term.KEY_UP,
        u'j': term.KEY_DOWN,
        u'b': term.KEY_PGUP,
        u'f': term.KEY_PGDOWN,
        u' ': term.KEY_PGDOWN,
        u'G': term.KEY_END,
        u'0': term.KEY_HOME,
    }
    return seq, van


def say_something(term, session):
    """ Prompt user to post a 'oneliner' entry. """
    # use same line as previous prompt, clearing it first,
    echo(term.move_x(0) + term.clear_eol)
    colors = {'highlight': term.red_reverse}

    xpos = max((term.width / 2) - ((MAX_MSGLEN / 2) + 3), 0)
    echo(term.move_x(xpos))
    if MIN_ELAPSED:
        lastliner = session.user.get('lastliner', 0)
        if lastliner and time.time() - lastliner > MIN_ELAPSED:
            echo(term.red_bold("You've said enough already!"))
            term.inkey(timeout=1)
            echo(term.move_x(0))
            # only re-draw prompt (user canceled)
            return False

    echo(term.move_x(xpos))
    echo(term.red('say ') + term.bold_red('>') + u' ')
    inp = LineEditor(MAX_MSGLEN, colors=colors).read()
    if not inp or not len(inp.strip()):
        # canceled
        echo(term.move_x(0) + term.clear_eol)
        return False

    echo(term.move_x(xpos) + term.clear_eol)
    session.user['lastliner'] = time.time()

    # optionally post to shroo.ms 'global oneliners'
    if shroo_ms_enabled:
        echo(term.red_reverse('Burning, please wait ...'))
        return post_shroo_ms(message=inp, username=session.user.handle)

    # or, post locally
    add_oneline(session=session, message=inp.strip())
    return True


def display_prompt(term, yloc):
    """ Display prompt of 'say something? [yn]'. """
    # display 'Say something? [yn]', with punctuation
    # in bold and 'n' underlined (default), then move
    # cursor backwards after '?' -- centered.
    echo(term.move(yloc, 0) + term.clear_eol)
    echo(term.center(
        u''.join(('Say something',
                  term.bold_white('? ['),
                  'y',
                  term.underline('N'),
                  term.bold_white(']'),
                  '\b\b\b\b\b'))
    ).rstrip())


def get_timeago(now, given_datestr):
    """ Return '3m' or some such as 'ago' time for given datestr. """
    return timeago(now - time.mktime(
        time.strptime(given_datestr, '%Y-%m-%d %H:%M:%S'))
    ).strip()


def generate_recent_oneliners(term, n_liners, offset):
    """ return string of all oneliners, its vert. hieght and adj. offset. """
    # generate a color palette
    palette = [getattr(term, _color) for _color in color_palette]

    # for relative 'time ago'
    now = time.time()

    # fetch all liners in database, sorted ascending by time
    oneliners = sorted(DBProxy('oneliner').values(), key=keysort_by_datetime)

    # decide the start/end by given offset, to allow paging, bounds check to
    # ensure that it does not scroll out of range
    offset = min(offset, len(oneliners))
    start, end = len(oneliners) - (n_liners + offset), len(oneliners) - offset
    if start < 0:
        offset += start
        start, end = 0, end - start
    if offset < 0:
        start, end = start + offset, end + offset
        offset = 0

    # build up one large text field; refresh is smoother when
    # all text is received as a single packet
    final_text_field = u''
    count = 0
    for count, oneliner in enumerate(oneliners[start:end]):
        _color = palette[count % len(palette)]
        ago = get_timeago(now, oneliner.get('timestamp'))
        alias = oneliner.get('alias', 'anonymous')
        bbsname = ('' if not shroo_ms_enabled else
                   oneliner.get('bbsname', 'untergrund'))
        content = (oneliner.get('oneliner', u''))
        max_msglen = MAX_MSGLEN
        if len(alias) > username_max_length:
            max_msglen -= (len(alias) - username_max_length)
        for _ in range(10):
            left_part = u'{0}: {1} '.format(
                alias.rjust(username_max_length),
                term.ljust(decode_pipe(content[:max_msglen]), max_msglen)
            )
            right_part = '{0} {1}'.format(_color(bbsname.rjust(4)), ago)
            txt_field = left_part + right_part
            if term.length(txt_field) < term.width:
                break
            elif term.length(left_part) < term.width:
                txt_field = left_part
                break
            max_msglen -= 2
        final_text_field = u''.join((
            final_text_field,
            term.move_x(max(0, (term.width / 2) - 45)),
            txt_field,
            term.clear_eol,
            u'\r\n')
        )

    # return text, vertical height, and adjusted offset
    return final_text_field, count, offset


def display_oneliners(term, top_margin, offset):
    """ Display recent oneliners, return bottom margin and adj. offset. """
    _padding = 3
    n_liners = term.height - _padding - top_margin
    txt, count, offset = generate_recent_oneliners(term, n_liners, offset)
    echo(u''.join((term.move(top_margin + 1, 0),
                   txt, u'\r\n')))
    bot_margin = top_margin + count + _padding
    return bot_margin, offset


def do_prompt(term, session):
    thread = None
    if shroo_ms_enabled:
        # fetch shroo-ms api oneliners
        thread = FetchUpdatesShrooMs()
        thread.start()
    dirty = -1
    top_margin = bot_margin = 0
    offset = 0
    do_quit = False
    while not do_quit:
        if dirty == -1:
            # re-display entire screen on-load, only. there
            # should never be any need to re-draw the art here-forward.
            top_margin = display_banner(art_file, encoding=art_encoding)
            echo(u'\r\n')
            top_margin += 1
            dirty = 1

        if dirty == 1:
            # re-display all oneliners on update, or dirty == 1
            # if any shroo-ms oneliners were received, merge them
            # into the database from the main thread.
            if thread is not None and not thread.is_alive():
                if thread.new_content:
                    echo(term.move_x(0) + term.clear_eol)
                    echo(term.center(term.bold_red('This just in!')).rstrip())
                    do_merge_shroo_ms(thread.new_content)
                thread = None

            with term.hidden_cursor():
                bot_margin, offset = display_oneliners(
                    term, top_margin, offset)
            dirty = 1

        if dirty:
            # always re-prompt on any dirty flag
            session.activity = u'Viewing Oneliners'
            display_prompt(term, yloc=bot_margin)
            dirty = 0

        event, data = session.read_events(
            ('input', 'oneliner', 'refresh'))
        if event == 'refresh':
            dirty = -1
            continue
        elif event == 'oneliner':
            dirty = 1
            continue
        elif event == 'input':
            session.buffer_input(data, pushback=True)

            inp = term.inkey(0)
            while inp:
                if inp.lower() in (u'y',):
                    # say something, refresh after
                    echo(inp)
                    say_retval = say_something(term, session)
                    if isinstance(say_retval, FetchUpdatesShrooMs):
                        # a rather strange hack, we track the thread of
                        # api calls so that we can merge its final content
                        # into our local database. See __init__ for
                        # bug id and description
                        thread = say_retval
                    elif say_retval:
                        # user has said something to the local database,
                        # refresh it so that they can beam with pride ...
                        dirty = 1
                        break

                    # only redraw prompt (user canceled)
                    dirty = 2
                elif inp.lower() in (u'n', u'q', u'\r', u'\n'):
                    echo(inp + u'\r\n')
                    do_quit = True
                    break

                elif len(inp):
                    # maybe scroll, really quite convoluted ...
                    height = bot_margin - top_margin
                    sequence_keymap, keymap = get_keymap(term, offset, height)
                    if ((inp.is_sequence and inp.code in sequence_keymap) or
                            inp in keymap):
                        _noff = sequence_keymap.get(
                            keymap.get(inp, None), offset)
                        n_offset = sequence_keymap.get(
                            inp.code, _noff)
                        if n_offset != offset:
                            offset = n_offset
                            dirty = 1
                inp = term.inkey(0)


def main():
    """ Script entry point. """
    session, term = getsession(), getterminal()

    echo(u'\r\n')

    # set syncterm font, if any
    if syncterm_font and term.kind.startswith('ansi'):
        echo(syncterm_setfont(syncterm_font))
        echo(term.move_x(0) + term.clear_eol)

    do_prompt(term, session)



""" File transfer routines for x/84. """

# local imports
from x84.bbs.session import getsession

# 3rd party
import xmodem


def send_modem(stream, protocol='xmodem1k', retry=16, timeout=30,
               callback=None):
    """
    Send a file using 'xmodem1k' or 'xmodem' protocol.

    Currently, these are the only protocols supported.  Returns ``True`` upon
    successful transmission, otherwise ``False``.

    :param stream: The file-like stream object to send data from.
    :param int retry: The maximum number of times to try to resend a failed
                      packet before failing.
    :param int timeout: seconds to elapse for response before failing.
    :param callable callback: Reference to a callback function that has the
           following signature. This is useful for getting
           status updates while a transfer is underway::

               def callback(total_count, success_count, error_count)
    """
    # get protocol implementation class
    supported_protocols = ('xmodem', 'xmodem1k')
    assert protocol in supported_protocols, (protocol, supported_protocols)
    Modem = {
        'xmodem': xmodem.XMODEM,
        'xmodem1k': xmodem.XMODEM1k,
    }[protocol]

    # the session's 'input' event buffer is used for receiving
    # transmissions.  It arrives in raw bytes, and session.write
    # is used, sending "unicode" data as encoding iso8859-1.
    session = getsession()

    def getc(size, timeout=10):
        """ Callback function for (X)Modem interface. """
        val = b''
        while len(val) < size:
            next_val = session.read_event('input', timeout=timeout)
            if next_val is None:
                break
            val += next_val
        if len(val) > size:
            session.buffer_input(val[size:])
        return val[:size] or None

    def putc(data, timeout=10):
        # pylint: disable=W0613
        #         Unused argument 'timeout'
        """ Callback function for (X)Modem interface. """
        session.write(data.decode('iso8859-1'), 'iso8859-1')

    modem = Modem(getc, putc)
    return modem.send(stream=stream, retry=retry, timeout=timeout,
                      quiet=True, callback=callback)


def recv_modem(stream, protocol='xmodem1k', retry=16, timeout=30):
    """
    Receive a file using 'xmodem1k' or 'xmodem' protocol into ``stream``.

    Currently, these are the only protocols supported.  Returns ``True`` upon
    successful transmission, otherwise ``False``.

    :param stream: The file-like stream object to send data from.
    :param int retry: The maximum number of times to try to resend a failed
                      packet before failing.
    :param int timeout: seconds to elapse for response before failing.
    """
    # get protocol implementation class
    supported_protocols = ('xmodem', 'xmodem1k')
    assert protocol in supported_protocols, (protocol, supported_protocols)
    Modem = {
        'xmodem': xmodem.XMODEM,
        'xmodem1k': xmodem.XMODEM1k,
    }[protocol]

    # the session's 'input' event buffer is used for receiving
    # transmissions.  It arrives in raw bytes, and session.write
    # is used, sending "unicode" data as encoding iso8859-1.
    session = getsession()

    def getc(size, timeout=10):
        """ Callback function for (X)Modem interface. """
        val = b''
        while len(val) < size:
            next_val = session.read_event('input', timeout=timeout)
            if next_val is None:
                break
            val += next_val
        if len(val) > size:
            session.buffer_input(val[size:])
        return val[:size] or None

    def putc(data, timeout=10):
        """ Callback function for (X)Modem interface. """
        # pylint: disable=W0613
        #         Unused argument 'timeout'
        session.write(data.decode('iso8859-1'), 'iso8859-1')

    modem = Modem(getc, putc)
    return modem.recv(stream=stream, retry=retry,
                      timeout=timeout, quiet=True)



""" Voting booth script for x/84. """

from common import waitprompt, prompt_pager
from x84.bbs import getsession, getterminal, echo, LineEditor
from x84.bbs import DBProxy, showart, syncterm_setfont
import os

databasename = 'votingbooth'  # change this to use an alternative database file

__author__ = 'Hellbeard'
__version__ = 1.2

# -----------------------------------------------------------------------------------

def ynprompt():
    term = getterminal()
    echo(term.magenta + u' (' + term.cyan + u'yes' + term.magenta +
         u'/' + term.cyan + u'no' + term.magenta + u')' + term.white)
    while True:
        inp = term.inkey()
        if inp.lower() == u'y':
            yn = True
            echo(u' yes!')
            break
        if inp.lower() == u'n' or inp.lower() == u'q':
            yn = False
            echo(u' no')
            break
    return(yn)

# -----------------------------------------------------------------------------------

def query_question():
    term = getterminal()
    session = getsession()
    db = DBProxy(databasename)
    questions = []
    index = []
    uservotingdata = []
    questions = db['questions']
    index = db['index']

    # create a new database file if none exists
    if not session.user.handle in db:
        db[session.user.handle] = {}
    uservotingdata = db[session.user.handle]

    echo(term.clear() + term.blue(u'>>') + term.white(u'questions availible\r\n') +
         term.blue(u'-' * 21 + '\r\n\r\n'))

    text = ''
    for i in range(0, len(questions)):
        if (index[i], 0) in uservotingdata:
            text = text + term.green(u'*')
        text = text + u''.join(term.magenta + u'(' + term.cyan + str(i) + term.magenta + u') ' +
                              term.white + questions[i] + u'\r\n')
    text = text.splitlines()
    prompt_pager(content=text,
                 line_no=0,
                 colors={'highlight': term.cyan,
                         'lowlight': term.green,
                         },
                 width=term.width, breaker=None, end_prompt=False)
    echo(term.move_x(0) + term.bold_black(u'* = already voted\r\n\r\n'))

    while True:
        echo(term.move_x(
            0) + term.magenta(u'select one of the questions above or press enter to return: '))
        le = LineEditor(10)
        le.colors['highlight'] = term.cyan
        inp = le.read()
        if inp is not None and inp.isnumeric() and int(inp) < len(questions):
            return int(inp)
        else:
            # -1 in this case means that no valid option was chosen.. break
            # loop.
            return -1

# -----------------------------------------------------------------------------------


def list_results(questionnumber):
    term = getterminal()
    db = DBProxy(databasename)
    alternatives = {}
    questions = []
    results = []
    amount_of_alternatives = db['amount_of_alternatives']
    alternatives = db['alternatives']
    questions = db['questions']
    results = db['results']

    echo(term.clear())

    text = (term.white + questions[questionnumber] + u'\r\n' + term.blue +
            u'-' * len(questions[questionnumber]) + u'\r\n\r\n')

    # only display full statistics if the screen width is above 79 columns.
    if term.width > 79:
        text = text + (term.magenta + u'(alternatives)' + term.move_x(49) +
                       u'(votes)' + term.move_x(57) + u'(percentage)\r\n')
        totalvotes = 0.00
        for i in range(0, amount_of_alternatives[questionnumber]):
            totalvotes = totalvotes + results[(questionnumber, i)]
        for i in range(0, amount_of_alternatives[questionnumber]):
            if results[(questionnumber, i)] > 0:
                percentage = (results[(questionnumber, i)] / totalvotes) * 100
            else:
                percentage = 0
            staple = int(round(percentage / 5))
            text = text + u''.join(term.move_x(0) + term.white(alternatives[(questionnumber, i)]) + term.move_x(49) +
                                   term.cyan(str(results[(questionnumber, i)])) + u'  ' + term.cyan + str(int(percentage)) +
                                   term.cyan(u'%') + term.move_x(57) + term.cyan(u'[') +
                                   term.green('#' * staple) + term.move_x(78) + term.cyan(']'))
            if i != amount_of_alternatives[questionnumber]:
                text = text + u'\r\n'
    else:
        for i in range(0, amount_of_alternatives[questionnumber]):
            text = text + (term.white(str(alternatives[(questionnumber, i)])) + term.cyan(u' votes: ') +
                           term.magenta(str(results[(questionnumber, i)])) + u'\r\n')

    text = text.splitlines()
    prompt_pager(content=text,
                 line_no=0,
                 colors={'highlight': term.cyan,
                         'lowlight': term.green,
                         },
                 width=term.width, breaker=None, end_prompt=False)
    echo(term.move_x(0) + term.bold_black(u'* = already voted\r\n'))
    waitprompt(term)

# -----------------------------------------------------------------------------------

def vote(questionnumber):
    term = getterminal()
    session = getsession()
    db = DBProxy(databasename)
    questions = []
    amount_of_alternatives = []
    alternatives = {}
    results = {}
    index = []
    questions = db['questions']
    alternatives = db['alternatives']
    results = db['results']
    amount_of_alternatives = db['amount_of_alternatives']
    index = db['index']

    echo(term.clear() + term.white + questions[questionnumber] + u'\r\n' +
         term.blue(u'-' * len(questions[questionnumber])) + u'\r\n\r\n')
    text = ''
    for i in range(0, amount_of_alternatives[questionnumber]):
        text = text + (term.magenta + u'(' + term.cyan + str(i) + term.magenta + u') ' +
                       term.white + alternatives[(questionnumber, i)] + u'\r\n')

    text = text.splitlines()
    prompt_pager(content=text,
                 line_no=0,
                 colors={'highlight': term.cyan,
                         'lowlight': term.green,
                         },
                 width=term.width, breaker=None, end_prompt=False)
    echo(term.move_x(0) + term.magenta(u'(') + term.cyan(str(amount_of_alternatives[questionnumber])) +
         term.magenta(u') ') + term.bold_black(u'Add your own answer..\r\n\r\n'))

    while True:
        echo(term.move_x(0) + term.magenta(u'Your choice: '))
        le = LineEditor(10)
        le.colors['highlight'] = term.cyan
        inp = le.read()
        if inp is not None and inp.isnumeric() and int(
                inp) <= amount_of_alternatives[questionnumber]:

            # create database for user if the user hasn't made any votes
            if session.user.handle not in db:
                db[session.user.handle] = {}

            uservotingdata = {}
            uservotingdata = db[session.user.handle]

            # if user wants to create an own alternative..
            if int(inp) == amount_of_alternatives[questionnumber]:
                echo(term.clear + term.red + u'\r\nPress enter to abort. ' +
                     term.move(0, 0) + term.white(u'Your answer: '))
                le = LineEditor(48)
                new_alternative = le.read()
                if new_alternative == '' or new_alternative == None:
                    return
                results[(questionnumber, int(inp))] = 0  # init..
                # init..
                alternatives[(questionnumber, int(inp))] = new_alternative
                amount_of_alternatives[
                    questionnumber] = amount_of_alternatives[questionnumber] + 1
                db['alternatives'] = alternatives
                db['amount_of_alternatives'] = amount_of_alternatives

            # if the user has voted on this question before..
            if (index[questionnumber], 0) in uservotingdata:
                temp2 = uservotingdata[(index[questionnumber], 0)]
                results[(questionnumber, temp2)] = results[
                    (questionnumber, temp2)] - 1  # remove the old vote
                results[(questionnumber, int(inp))] = results[
                    (questionnumber, int(inp))] + 1
                uservotingdata[(index[questionnumber], 0)] = int(inp)
            else:
                uservotingdata[(index[questionnumber], 0)] = int(inp)
                results[(questionnumber, int(inp))] = results[
                    (questionnumber, int(inp))] + 1

            uservotingdata[(index[questionnumber], 0)] = int(inp)

            echo(term.green(u'\r\nyour vote has been noted, thanks..'))
            term.inkey(2)
            db['results'] = results
            db[session.user.handle] = uservotingdata
            list_results(questionnumber)
            return

# -----------------------------------------------------------------------------------

def add_question():
    term = getterminal()
    db = DBProxy(databasename)
    questions = []
    amount_of_alternatives = []
    index = {}
    alternatives = {}
    results = {}
    index = db['index']
    questions = db['questions']
    alternatives = db['alternatives']
    results = db['results']
    amount_of_alternatives = db['amount_of_alternatives']
    amount_of_questions = len(questions)

    echo(term.clear + term.white + u'\r\nQuestion: ')
    le = LineEditor(65)
    new_question = le.read()
    if new_question == '' or new_question == None:
        return

    echo(term.bold_black(u'\r\n\r\nLeave a blank line when you are finished..'))
    new_amount = 0
    while True:
        echo(term.normal + term.white + u'\r\nchoice ' +
             term.red + str(new_amount) + term.white + u': ')
        le = LineEditor(48)
        alternatives[(amount_of_questions, new_amount)] = le.read()
        if alternatives[(amount_of_questions, new_amount)] == '' or alternatives[(amount_of_questions, new_amount)] == None :
            break
        else:
            results[(amount_of_questions, new_amount)] = 0
            new_amount = new_amount + 1

    if new_amount > 0:
        echo(term.white(u'\r\n\r\nSave this voting question?'))
        answer = ynprompt()
        if answer == 1:
            questions.append(new_question)
            amount_of_alternatives.append(new_amount)

            indexcounter = db['indexcounter']
            indexcounter = indexcounter + 1
            index.append(str(indexcounter))

            db['indexcounter'] = indexcounter
            db['index'] = index
            db['questions'] = questions
            db['amount_of_alternatives'] = amount_of_alternatives
            db['results'] = results
            db['amount_of_questions'] = amount_of_questions
            db['alternatives'] = alternatives

            waitprompt(term)

# -----------------------------------------------------------------------------------

def delete_question(questionnumber):
    term = getterminal()
    db = DBProxy(databasename)
    alternatives = {}
    questions = []
    results = {}
    amount_of_alternatives = []
    questions = db['questions']
    results = db['results']
    amount_of_alternatives = db['amount_of_alternatives']
    alternatives = db['alternatives']
    index = db['index']

    echo(term.clear + term.white(u'Delete the ') + term.magenta(u'(') + term.cyan(u'e') + term.magenta(u')') +
         term.white(u'ntire question or delete single ') + term.magenta(u'(') + term.cyan(u'a') + term.magenta(u')') +
         term.white(u'lternatives?') + u'\r\n\r\n' + term.magenta(u'command: '))

    le = LineEditor(10)
    le.colors['highlight'] = term.cyan
    inp = le.read()
    # makes the input indifferent to wheter you used lower case when typing in
    # a command or not..
    inp = (inp or u'').lower()

    if inp == u'a':  # delete answer alternative..
        echo(term.clear)
        echo(term.white + questions[questionnumber] + term.move_x(max(0, term.width - 12)) +
             u' index: ' + str(index[questionnumber]) + u'\r\n\r\n')
        for i in range(0, amount_of_alternatives[questionnumber]):
            echo(term.cyan(str(i) + u'. ') +
                 term.white(alternatives[(questionnumber, i)]) + u'\r\n')

        echo(term.magenta(u'\r\nSelect a number. Enter to abort: '))

        le = LineEditor(10)
        le.colors['highlight'] = term.cyan
        inp2 = le.read()

        if inp2 is not None and inp2.isnumeric() and int(
                inp2) < amount_of_alternatives[questionnumber]:
            if int(inp2) + 1 < amount_of_alternatives[questionnumber]:
                for i in range(
                        int(inp2), amount_of_alternatives[questionnumber] - 1):
                    alternatives[(questionnumber, i)] = alternatives[
                        (questionnumber, i + 1)]
                    results[(questionnumber, i)] = results[
                        (questionnumber, i + 1)]
        else:
            return
        amount_of_alternatives[questionnumber] -= 1

    elif inp == u'e':  # delete entire question..
        if questionnumber + 1 < len(questions):
            for i in range(questionnumber, len(questions) - 1):
                questions[i] = questions[i + 1]
                amount_of_alternatives[i] = amount_of_alternatives[i + 1]
                index[(i)] = index[(i + 1)]
                for i2 in range(0, amount_of_alternatives[i + 1]):
                    alternatives[(i, i2)] = alternatives[(i + 1, i2)]
                    results[(i, i2)] = results[(i + 1, i2)]
        del questions[-1]
        del amount_of_alternatives[-1]
        del index[-1]
    else:
        return

    db['index'] = index
    db['questions'] = questions
    db['amount_of_alternatives'] = amount_of_alternatives
    db['results'] = results
    db['alternatives'] = alternatives
    return

# -----------------------------------------------------------------------------------

def generate_database():  # generates a database file with a generic question.
    db = DBProxy(databasename)

    index = []
    index.append(0)
    indexcounter = 0

    questions = []
    questions.append('Which is your prefered BBS software?')

    alternatives = {}
    alternatives[(0, 0)] = 'X/84'
    alternatives[(0, 1)] = 'Daydream'
    alternatives[(0, 2)] = 'Mystic'
    alternatives[(0, 3)] = 'Synchronet'

    results = {}
    results[(0, 0)] = 0
    results[(0, 1)] = 0
    results[(0, 2)] = 0
    results[(0, 3)] = 0

    amount_of_alternatives = []
    # this is the only list/dict that is not zerobased..
    amount_of_alternatives.append(4)

    db['indexcounter'] = indexcounter
    db['index'] = index
    db['amount_of_alternatives'] = amount_of_alternatives
    db['alternatives'] = alternatives
    db['results'] = results
    db['questions'] = questions

# -----------------------------------------------------------------------------------

def main():
    session = getsession()
    session.activity = u'hanging out in voting script'
    term = getterminal()
    echo(syncterm_setfont('topaz'))

    db = DBProxy(databasename)
    if 'questions' not in db:
        generate_database()

    while True:
        # clears the screen and displays the vote art header
        echo(term.clear())
        for line in showart(
                os.path.join(os.path.dirname(__file__), 'art', 'vote.ans'), 'cp437'):
            echo(term.cyan + term.move_x(max(0, (term.width / 2) - 40)) + line)

        if 'sysop' in session.user.groups:
            spacing = 1
        else:
            spacing = 7
            echo(' ')
        echo(term.magenta(u'\n (') + term.cyan(u'r') + term.magenta(u')') +
             term.white(u'esults') + ' ' * spacing +
             term.magenta(u'(') + term.cyan(u'v') + term.magenta(u')') +
             term.white(u'ote on a question') + u' ' * spacing +
             term.magenta(u'(') + term.cyan(u'a') + term.magenta(u')') +
             term.white(u'dd a new question') + u' ' * spacing)
        if 'sysop' in session.user.groups:
            echo(term.magenta(u'(') + term.cyan(u'd') + term.magenta(u')') +
                 term.white(u'elete a question') + u' ' * spacing)
        echo(term.magenta(u'(') + term.cyan(u'q') + term.magenta(u')') + term.white(u'uit') +
             term.magenta(u'\r\n\r\nx/84 voting booth command: '))
        le = LineEditor(10)
        le.colors['highlight'] = term.cyan
        inp = le.read()
        # makes the input indifferent to wheter you used lower case when typing
        # in a command or not..
        inp = (inp or u'').lower()

        if 'sysop' in session.user.groups and inp == u'd':
            while True:
                questionnumber = query_question()
                if questionnumber == -1:
                    break
                delete_question(questionnumber)
        elif inp == u'r':
            while True:
                questionnumber = query_question()
                if questionnumber == -1:
                    break
                list_results(questionnumber)
        elif inp == u'v':
            while True:
                questionnumber = query_question()
                if questionnumber == -1:
                    break
                vote(questionnumber)
        elif inp == u'a':
            add_question()
        elif inp == u'q':
            return
        else:
            # if no valid key is pressed then do some ami/x esthetics.
            echo(term.red(u'\r\nNo such command. Try again.\r\n'))
            waitprompt(term)



""" top-level scripting module for x/84. """
# local side-effect producing imports
# (encodings such as 'cp437_art' become registered)
__import__('encodings.aliases')
__import__('x84.encodings')

# local/exported at top-level 'from bbs import ...'
from x84.bbs.ansiwin import AnsiWindow
from x84.bbs.dbproxy import DBProxy
from x84.bbs.door import Door, DOSDoor, Dropfile
from x84.bbs.editor import LineEditor, ScrollingEditor
from x84.bbs.exception import Disconnected, Goto
from x84.bbs.ini import get_ini
from x84.bbs.lightbar import Lightbar
from x84.bbs.modem import send_modem, recv_modem
from x84.bbs.msgbase import list_msgs, get_msg, list_tags, Msg, list_privmsgs
from x84.bbs.output import (echo, timeago, encode_pipe, decode_pipe,
                            syncterm_setfont, showart, ropen,
                            from_cp437,  # deprecated in v2.0
                            )
from x84.bbs.pager import Pager
from x84.bbs.script_def import Script
from x84.bbs.selector import Selector
from x84.bbs.session import (getsession, getterminal,
                             goto, disconnect, gosub,
                             getch,      # deprecated in v2.1
                             )
from x84.bbs.userbase import list_users, get_user, find_user, User, Group

# the scripting API is generally defined by this __all__ attribute, but
# the real purpose of __all__ is defining what gets placed into a caller's
# namespace when using statement `from x84.bbs import *`
__all__ = ('list_users', 'get_user', 'find_user', 'User', 'Group', 'list_msgs',
           'get_msg', 'list_tags', 'Msg', 'LineEditor', 'ScrollingEditor',
           'echo', 'timeago', 'AnsiWindow', 'Selector', 'Disconnected', 'Goto',
           'Lightbar', 'from_cp437', 'DBProxy', 'Pager', 'Door', 'DOSDoor',
           'goto', 'disconnect', 'getsession', 'getterminal', 'getch', 'gosub',
           'ropen', 'showart', 'Dropfile', 'encode_pipe',
           'decode_pipe', 'syncterm_setfont', 'get_ini', 'send_modem',
           'recv_modem', 'Script', 'list_privmsgs',
           )



from __future__ import absolute_import

import codecs
import logging
import re

_cache = {}
_aliases = {}

logger = logging.getLogger(__name__)


def normalize_encoding(encoding):
    return re.sub(
        r'[^\w_-]', '', encoding.lower()
    ).replace('-', '_')


def search_function(encoding):
    try:
        return _cache[encoding]
    except KeyError:
        pass

    encoding = _aliases.get(encoding, encoding)
    mod = None
    try:
        mod = __import__('x84.encodings.' + encoding, fromlist=['*'], level=0)
    except ImportError:
        pass

    try:
        getregentry = mod.getregentry
    except AttributeError:
        mod = None

    if mod is None:
        _cache[encoding] = None
        return None

    _cache[encoding] = getregentry()

    try:
        codecaliases = mod.getaliases()
    except AttributeError:
        pass
    else:
        for alias in codecaliases:
            if alias not in _aliases:
                _aliases[alias] = _cache[encoding]

    return _cache[encoding]


codecs.register(search_function)


# Now to initialize all locally available codecs:
for encoding in ('amiga', 'atarist', 'cp437_art', 'cp437'):
    ''.decode(encoding)



"""
Message area for x/84.

This script provides an interface to check for new
messages, subscribe to and browse tags and networks,
find all, or unread messages, or all messages since
last login.

It determines a set of message-ids that are then
forwarded to the message browser interface.
"""
# std imports
import collections
import datetime
import fnmatch
import difflib

# local
from x84.bbs import (
    syncterm_setfont,
    ScrollingEditor,
    list_privmsgs,
    decode_pipe,
    getterminal,
    getsession,
    LineEditor,
    list_users,
    list_msgs,
    list_tags,
    get_ini,
    get_msg,
    timeago,
    DBProxy,
    gosub,
    echo,
    Msg,
)
from common import (
    render_menu_entries,
    show_description,
    display_banner,
    display_prompt,
    prompt_pager,
    waitprompt,
)

# 3rd-party
import dateutil

TIME_FMT = '%A %b-%d, %Y at %r UTC'

#: banner art displayed in main()
art_file = get_ini(
    section='msgarea', key='art_file'
) or 'art/hx-msg.ans'

#: character encoding of banner art
art_encoding = get_ini(
    section='msgarea', key='art_encoding'
) or 'cp437'

#: preferred fontset for SyncTerm emulator
syncterm_font = get_ini(
    section='msgarea', key='syncterm_font'
) or 'topaz'

#: When set False, menu items are not colorized and render much
#: faster on slower systems (such as raspberry pi).
colored_menu_items = get_ini(
    section='msgarea', key='colored_menu_items', getter='getboolean'
) or True

#: color used for description text
color_text = get_ini(
    section='msgarea', key='color_text'
) or 'white'

#: color used for menu key entries
color_highlight = get_ini(
    section='msgarea', key='color_highlight'
) or 'bold_magenta'

#: color used for prompt
color_backlight = get_ini(
    section='msgarea', key='color_prompt',
) or 'magenta_reverse'

#: color used for brackets ``[`` and ``]``
color_lowlight = get_ini(
    section='msgarea', key='color_lowlight'
) or 'bold_black'

#: maximum length of user handles
username_max_length = get_ini(
    section='nua', key='max_user', getter='getint'
) or 10

subject_max_length = get_ini(
    section='msg', key='max_subject', getter='getint'
) or 40


def get_menu(messages):
    """ Return list of menu items by given dict ``messages``. """
    MenuItem = collections.namedtuple('MenuItem', ['inp_key', 'text'])
    items = []
    if messages['new']:
        items.extend([
            MenuItem(u'n', u'new ({0})'.format(len(messages['new']))),
            MenuItem(u'm', u'mark all read'),
        ])
    if messages['all']:
        items.append(
            MenuItem(u'a', u'all ({0})'.format(len(messages['all'])))
        )
    if messages['private']:
        items.append(
            MenuItem(u'v', u'private ({0})'.format(len(messages['private'])))
        )
    items.extend([
        MenuItem(u'p', u'post public'),
        MenuItem(u'w', u'write private'),
        MenuItem(u'c', u'change area'),
        MenuItem(u'?', u'help'),
        MenuItem(u'q', u'quit'),
    ])
    return items


def do_mark_as_read(session, message_indicies):
    """ Mark all given messages read. """
    session.user['readmsgs'] = (
        session.user.get('readmsgs', set()) | set(message_indicies)
    )


def get_messages_by_subscription(session, subscription):
    all_tags = list_tags()
    messages = {'all': set(), 'new': set()}
    messages_bytag = {}
    messages_read = session.user.get('readmsgs', set())

    # now occlude all private messages :)
    all_private = list_privmsgs(None)

    # this looks like perl code
    for tag_pattern in subscription:
        messages_bytag[tag_pattern] = collections.defaultdict(set)
        for tag_match in fnmatch.filter(all_tags, tag_pattern):
            msg_indicies = list_msgs(tags=(tag_match,))
            messages['all'].update(msg_indicies - all_private)
            messages_bytag[tag_pattern]['all'].update(msg_indicies - all_private)
        messages_bytag[tag_pattern]['new'] = (
            messages_bytag[tag_pattern]['all'] - messages_read)

    # and make a list of only our own
    messages['private'] = list_privmsgs(session.user.handle)

    # and calculate 'new' messages
    messages['new'] = (messages['all'] | messages['private']) - messages_read

    return messages, messages_bytag


def describe_message_area(term, subscription, messages_bytags, colors):
    get_num = lambda lookup, tag_pattern, grp: len(lookup[tag_pattern][grp])
    return u''.join((
        colors['highlight'](u'msgarea: '),
        colors['text'](u', ').join((
            u''.join((
                quote(tag_pattern, colors),
                u'({num_new}/{num_all})'.format(
                    num_new=get_num(messages_bytags, tag_pattern, 'new'),
                    num_all=get_num(messages_bytags, tag_pattern, 'all'))
            )) for tag_pattern in subscription)),
        u'\r\n\r\n',
    ))


def validate_tag_patterns(tag_patterns):
    all_tags = list_tags() or set([u'public'])
    removed = []
    for tag_pattern in set(tag_patterns):
        if not fnmatch.filter(all_tags, tag_pattern):
            removed.append(tag_pattern)
            tag_patterns.remove(tag_pattern)
    return removed, tag_patterns


def quote(txt, colors):
    return u''.join(((u'"'), colors['highlight'](txt), (u'"')))


def do_describe_available_tags(term, colors):
    sorted_tags = sorted([(len(list_msgs(tags=(tag,))), tag)
                          for tag in list_tags() or [u'public']
                          ], reverse=True)
    decorated_tags = [
        colors['text'](tag) +
        colors['lowlight']('({0})'.format(num_msgs))
        for num_msgs, tag in sorted_tags]

    description = u''.join((
        colors['highlight'](u'available tags'), ': ',
        colors['text'](u', ').join(decorated_tags),
        colors['text'](u'.'),
    ))
    return show_description(term, description, color=None)


def get_network_tag_description(term, colors):
    """ Return description text of message networks, if any. """
    server_tags = get_ini('msg', 'server_tags', split=True)
    network_tags = get_ini('msg', 'network_tags', split=True)

    if not (network_tags or server_tags):
        return u''
    return u''.join((
        u'\r\n\r\n',
        colors['text'](u'This board participates in intra-bbs '
                       'messaging, '),
        u''.join((
            colors['text'](u'hosting network messages by tag '),
            u', '.join(quote(tag, colors) for tag in server_tags),
        )) if server_tags else u'',
        (colors['text'](
            u' and ') if (server_tags and network_tags) else u''),
        u''.join((
            colors['text'](u'participating in network messages by tag '),
            u', '.join(quote(tag, colors) for tag in network_tags),
        )) if network_tags else u'',
        u'.',
    ))


def do_describe_message_system(term, colors):
    """ Display help text about message tagging. """

    def describe_group_tags():
        groups = getsession().user.groups
        if not groups:
            return u''

        return u''.join((
            u'\r\n\r\n',
            colors['text'](
                u'Finally, private messages may be shared among groups.  You '
                u'may post messages to any group you are a member of: '),
            colors['text'](
                u', '.join(quote(grp, colors) for grp in groups)),
            colors['text'](u'.')
        ))

    description = u''.join((
        u'\r\n',
        colors['text'](
            u'You can think of tags as a system of providing context to any '
            u'message stored on this system.  A tag might provide the '
            u'general label of the topic of conversation, which may be '
            u'subscribed to.  For example, '),
        quote(u'python', colors),
        colors['text'](
            u' may be used for topics related to the python programming '
            u'language.  This is similar to flicker or gmail tags, or '
            u'hashtags.  Public messages are always tagged '),
        quote(u'public', colors),
        colors['text'](u'.  '),
        get_network_tag_description(term, colors),
        u'\r\n\r\n',
        colors['text'](
            u'Furthermore, glob expressions may be used such as '),
        quote(u'*', colors),
        u' ',
        colors['text']('for all messages, or expression '),
        quote(u'lang-*', colors),
        u' ',
        colors['text']('might subscribe to both '),
        quote(u'lang-python', colors),
        colors['text'](u' and '),
        quote(u'lang-go', colors),
        colors['text'](u'.'),
        describe_group_tags(),
    ))
    return show_description(term, description, color=None)


def prompt_subscription(session, term, yloc, subscription, colors):
    """
    This function is called to assign a new set of subscription
    tags for a user.  If escape is pressed, the existing value
    is used, or '*' is used if not previously set.

    This should be called for first-time users, and optionally
    at any later time to change a subscription.
    """

    if session.user.get('msg_subscription', None) is None:
        # force-display introductory description for first-time users.
        yloc += do_describe_message_system(term, colors)
        echo(u'\r\n\r\n')
        yloc += 2

    # remind ourselves of all available tags
    yloc += do_describe_available_tags(term, colors) + 2

    # for small screens, scroll and leave room for prompt & errors
    if yloc > term.height + 3:
        echo(u'\r\n' * 3)
        yloc = term.height - 3

    # and prompt for setting of message tags
    xloc = max(0, (term.width // 2) - 40)
    input_prefix = u':: subscription tags:'
    echo(u''.join((term.move(yloc, xloc), input_prefix)))

    xloc += len(input_prefix)
    wide = min(40, (term.width - xloc - 2))

    while True:
        editor = ScrollingEditor(xloc=xloc, yloc=yloc - 1, width=wide,
                                 colors={'highlight': colors['backlight']},
                                 content=u', '.join(subscription),
                                 max_length=100)

        # Prompt for and evaluate the given input, splitting by comma,
        # removing any empty items, and defaulting to ['*'] on escape.
        inp = editor.read() or u''
        subscription = filter(None, set(map(unicode.strip, inp.split(',')))
                              ) or set([u'*'])

        # Then, reduce to only validate tag patterns, tracking those
        # that do not match any known tags, and display a warning and
        # re-prompt if any are removed.
        removed, subscription = validate_tag_patterns(subscription)

        # clear existing warning, if any
        echo(u''.join((term.normal, u'\r\n\r\n', term.clear_eos)))
        if removed:
            # and display any unmatched tags as a warning, re-prompt
            txt = ''.join((
                term.bold_red(u"The following patterns are not matched: "),
                u', '.join(removed)))
            show_description(term, txt, color=None)
            continue

        # otherwise everything is fine,
        # return new subscription set
        return subscription


def allow_tag(session, idx):
    """
    Whether user is allowed to tag a message.

    :rtype: bool

    A user can tag a message if the given session's user is:

    * the message author or recipient.
    * a member of sysop or moderator group.
    * a member of any existing tag-matching user group.
    """
    moderated = get_ini('msg', 'moderated_tags', getter='getboolean')
    tag_moderators = set(get_ini('msg', 'tag_moderators', split=True))
    if not moderated and 'sysop' in session.user.groups:
        return True

    elif moderated and (tag_moderators | session.user.groups):
        # tags are moderated, but user is one of the moderator groups
        return True

    msg = get_msg(idx)
    if session.user.handle in (msg.recipient, msg.author):
        return True

    for tag in msg.tags:
        if tag in session.user.groups:
            return True
    return False


def create_reply_message(session, idx):
    """ Given a message ``idx``, create and return a replying message. """
    parent_msg = get_msg(idx)
    msg = Msg()
    msg.parent = parent_msg.idx

    # flip from/to
    msg.recipient = parent_msg.author
    msg.author = session.user.handle

    # quote message body
    msg.body = quote_body(parent_msg)

    # duplicate subject and tags
    msg.subject = parent_msg.subject
    msg.tags = parent_msg.tags

    return msg


def quote_body(msg):
    """ Return quoted body of given message, ``msg``. """
    # chose a header separator, we iterate through each one, finding any that
    # may already be used, we use whichever is used least, or the one following
    # the last used.  The given quotesep_chars is modeled after
    # reStructuredText.
    quotesep_len = 60
    quotesep_chars = u"=-`'~*+^"
    # assign quotesep_char as quotesep_chars index ([1]) of
    # lowest-order/fewest uses of quote character([0]).
    quotesep_char = quotesep_chars[sorted(
        [(msg.body.count(given_sep * quotesep_len),
          quotesep_chars.index(given_sep))
         for given_sep in quotesep_chars])[0][1]]
    txt_sent = msg.stime.replace(
        tzinfo=dateutil.tz.tzlocal()
    ).astimezone(dateutil.tz.tzutc()).strftime(TIME_FMT)
    txt_who = u'|13{0}|07'.format(msg.author)
    return (u'On {txt_sent}, {txt_who} wrote:\r\n'
            u'{quotesep}\r\n{msg.body}\r\n{quotesep}'
            .format(txt_sent=txt_sent, txt_who=txt_who,
                    quotesep=(quotesep_char * quotesep_len),
                    msg=msg))


def display_message(session, term, msg_index, colors):
    """ Format message of index ``idx``. """
    color_handle = lambda handle: (
        colors['highlight'](handle)
        if handle == session.user.handle
        else handle)
    msg = get_msg(msg_index)
    txt_sent = msg.stime.replace(
        tzinfo=dateutil.tz.tzlocal()
    ).astimezone(dateutil.tz.tzutc()).strftime(TIME_FMT)
    txt_sentago = colors['highlight'](
        timeago((datetime.datetime.now() - msg.stime)
                .total_seconds()).strip())
    txt_to = color_handle(msg.recipient)
    txt_private = (colors['highlight'](' (private)')
                   if not 'public' in msg.tags else u'')
    txt_from = color_handle(msg.author)
    txt_tags = u', '.join((quote(tag, colors)
                           for tag in msg.tags))
    txt_subject = colors['highlight'](msg.subject)
    txt_body = decode_pipe(msg.body)
    txt_breaker = ('-' if session.encoding == 'ansi' else u'\u2500'
                   ) * min(80, term.width)
    msg_txt = (
        u'\r\n{txt_breaker}\r\n'
        u'   from: {txt_from}\r\n'
        u'     to: {txt_to}{txt_private}\r\n'
        u'   sent: {txt_sent} ({txt_sentago} ago)\r\n'
        u'   tags: {txt_tags}\r\n'
        u'subject: {txt_subject}\r\n'
        u'\r\n'
        u'{txt_body}\r\n'
        .format(txt_breaker=txt_breaker,
                txt_from=txt_from, txt_to=txt_to,
                txt_sent=txt_sent, txt_sentago=txt_sentago,
                txt_tags=txt_tags, txt_subject=txt_subject,
                txt_body=txt_body, txt_private=txt_private))

    do_mark_as_read(session, [msg_index])

    prompt_pager(content=msg_txt.splitlines(), line_no=0,
                 width=min(80, term.width),
                 colors=colors, breaker=u'- ', end_prompt=False,
                 break_long_words=True)


def can_delete(session):
    moderated = get_ini('msg', 'moderated_tags', getter='getboolean')
    tag_moderators = set(get_ini('msg', 'tag_moderators', split=True))
    return ('sysop' in session.user.groups or
            moderated and tag_moderators & session.user.groups)


def delete_message(msg):
    """ Experimental message delete! """
    # ! belongs as msg.delete() function !
    msg.recipient = u''
    msg.subject = u''
    msg.body = u''
    msg.children = set()
    msg.parent = None
    msg.tags = set()
    msg.save()
    with DBProxy('tags') as tag_db:
        for key, values in tag_db.items()[:]:
            if msg.idx in values:
                newvalue = values - set([msg.idx])
                if newvalue:
                    tag_db[key] = newvalue
                else:
                    # no more messages by this tag, delete it
                    del tag_db[key]
    with DBProxy('privmsg') as priv_db:
        for key, values in priv_db.items()[:]:
            if msg.idx in values:
                priv_db[key] = values - set([msg.idx])
    with DBProxy('msgbase') as msg_db:
        del msg_db['%d' % int(msg.idx)]


def do_reader_prompt(session, term, index, message_indices, colors):
    xpos = max(0, (term.width // 2) - (80 // 2))
    opts = []
    if index:
        opts += (('p', 'rev'),)
    if index < len(message_indices) - 1:
        opts += (('n', 'ext'),)
    if allow_tag(session, message_indices[index]):
        opts += (('e', 'dit tags'),)
    if can_delete(session):
        opts += (('D', 'elete'),)
    opts += (('r', 'eply'),)
    opts += (('q', 'uit'),)
    opts += (('idx', ''),)
    while True:
        echo(term.move_x(xpos))
        echo(u''.join((
            colors['lowlight'](u'['),
            colors['highlight'](str(index + 1)),
            u'/{0}'.format(len(message_indices)),
            colors['lowlight'](u']'),
            u' ',
            u', '.join((
                u''.join((colors['lowlight'](u'['),
                          colors['highlight'](key),
                          colors['lowlight'](u']'),
                          value
                          )) for key, value in opts)),
            u': ',
            term.clear_eol,
        )))
        width = max(2, len(str(len(message_indices))))
        inp = LineEditor(width, colors={'highlight': colors['backlight']}).read()
        if inp is None or inp.lower() == u'q':
            return None
        elif inp in (u'n', u''):
            # 'n'ext or return key
            echo(term.move_x(xpos) + term.clear_eol)
            if index == len(message_indices) - 1:
                # no more messages,
                return None
            return index + 1
        elif inp == u'p' and index > 0:
            # prev
            echo(term.move_x(xpos) + term.clear_eol)
            return index - 1
        elif inp == u'e' and allow_tag(session, message_indices[index]):
            msg = get_msg(message_indices[index])
            echo(u'\r\n')
            if prompt_tags(session=session, term=term, msg=msg,
                           colors=colors, public='public' in msg.tags):
                echo(u'\r\n')
                msg.save()
            return index
        elif inp == u'D' and can_delete(session):
            delete_message(msg=get_msg(message_indices[index]))
            return None
        elif inp == u'r':
            # write message reply
            msg = create_reply_message(session=session,
                                       idx=message_indices[index])
            if (
                    prompt_subject(
                        term=term, msg=msg, colors=colors
                    ) and prompt_body(
                        term=term, msg=msg, colors=colors
                    ) and prompt_tags(
                        session=session, term=term, msg=msg,
                        colors=colors, public='public' in msg.tags
                    )):
                do_send_message(session=session, term=term,
                                msg=msg, colors=colors)
            break
        else:
            # not a valid input option, is it a valid integer? (even '-1'!)
            try:
                val = int(inp)
            except ValueError:
                # some garbage; try again
                term.inkey(0.15)
                continue
            try:
                # allow a message index, (even pythonic '-1' for 'last')
                if val > 0:
                    # 1-based indexing
                    val -= 1
                nxt_idx = message_indices.index(message_indices[val])
                if nxt_idx != index:
                    echo(term.move_x(xpos) + term.clear_eol)
                    return nxt_idx
            except (IndexError, ValueError):
                # invalid index; try again
                term.inkey(0.15)
                continue


def read_messages(session, term, message_indices, colors):
    """ Read list of given messages. """
    index = 0
    while True:
        session.activity = ('reading msgs [{0}/{1}]'
                            .format(index + 1, len(message_indices)))
        display_message(session=session, term=term,
                        msg_index=message_indices[index],
                        colors=colors)
        index = do_reader_prompt(session=session, term=term, index=index,
                                 message_indices=message_indices,
                                 colors=colors)
        if index is None:
            break


def main(quick=False):
    """ Main procedure. """

    session, term = getsession(), getterminal()
    session.activity = 'checking for new messages'

    # set syncterm font, if any
    if term.kind.startswith('ansi'):
        echo(syncterm_setfont(syncterm_font))

    colors = dict(
        highlight=lambda txt: txt,
        lowlight=lambda txt: txt,
        backlight=lambda txt: txt,
        text=lambda txt: txt
    ) if not colored_menu_items else dict(
        highlight=getattr(term, color_highlight),
        lowlight=getattr(term, color_lowlight),
        backlight=getattr(term, color_backlight),
        text=getattr(term, color_text))

    yloc = top_margin = 0
    subscription = session.user.get('msg_subscription', [])
    dirty = 2

    while True:
        if dirty == 2:
            # display header art,
            yloc = display_banner(art_file, encoding=art_encoding, center=True)
            xloc = max(0, (term.width // 2) - 40)
            echo(u'\r\n')
            top_margin = yloc = (yloc + 1)

        elif dirty:
            echo(term.move(top_margin, 0) + term.normal + term.clear_eos)
            echo(term.move(top_margin, xloc))

        if dirty:

            if not subscription:
                # prompt the user for a tag subscription, and loop
                # back again when completed to re-draw and show new messages.
                subscription = session.user['msg_subscription'] = (
                    prompt_subscription(
                        session=session, term=term, yloc=top_margin,
                        subscription=subscription, colors=colors))
                continue

            messages, messages_bytags = get_messages_by_subscription(
                session, subscription)

            # When quick login ('y') selected in top.py, return immediately
            # when no new messages are matched any longer.
            if quick and not messages['new']:
                echo(term.move_x(xloc) + u'\r\nNo new messages.\r\n')
                return waitprompt(term)

            txt = describe_message_area(
                term=term, subscription=subscription,
                messages_bytags=messages_bytags, colors=colors)

            yloc = top_margin + show_description(
                term=term, description=txt, color=None,
                subsequent_indent=' ' * len('message area: '))

            echo(render_menu_entries(
                term=term, top_margin=yloc,
                menu_items=get_menu(messages),
                colors=colors, max_cols=2))
            echo(display_prompt(term=term, colors=colors))
            echo(colors['backlight'](u' \b'))
            dirty = False

        event, data = session.read_events(('refresh', 'newmsg', 'input'))

        if event == 'refresh':
            # screen resized, redraw.
            dirty = 2
            continue

        elif event == 'newmsg':
            # When a new message is sent, 'newmsg' event is broadcasted.
            session.flush_event('newmsg')
            nxt_msgs, nxt_bytags = get_messages_by_subscription(
                session, subscription)
            if nxt_msgs['new'] - messages['new']:
                # beep and re-display when a new message has arrived.
                echo(u'\b')
                messages, messages_bytags = nxt_msgs, nxt_bytags
                dirty = True
                continue

        elif event == 'input':
            # on input, block until carriage return
            session.buffer_input(data, pushback=True)
            given_inp = LineEditor(
                1, colors={'highlight': colors['backlight']}
            ).read()

            if given_inp is None:
                # escape/cancel
                continue

            inp = given_inp.strip()
            if inp.lower() in (u'n', 'a', 'v'):
                # read new/all/private messages
                message_indices = sorted(list(
                    {'n': messages['new'],
                     'a': messages['all'],
                     'v': messages['private'],
                     }[inp.lower()]))
                if message_indices:
                    dirty = 2
                    read_messages(session=session, term=term,
                                  message_indices=message_indices,
                                  colors=colors)
            elif inp.lower() == u'm' and messages['new']:
                # mark all messages as read
                dirty = 1
                do_mark_as_read(session, messages['new'])
            elif inp.lower() in (u'p', u'w'):
                # write new public/private message
                dirty = 2
                public = bool(inp.lower() == u'p')
                msg = Msg()
                if (
                        not prompt_recipient(
                            term=term, msg=msg,
                            colors=colors, public=public
                        ) or not prompt_subject(
                            term=term, msg=msg, colors=colors
                        ) or not prompt_body(
                            term=term, msg=msg, colors=colors
                        ) or not prompt_tags(
                            session=session, term=term, msg=msg,
                            colors=colors, public=public
                        )):
                    continue
                do_send_message(session=session, term=term,
                                msg=msg, colors=colors)
            elif inp.lower() == u'c':
                # prompt for new tag subscription (at next loop)
                subscription = []
                dirty = 1
            elif inp.lower() == u'?':
                # help
                echo(term.move(top_margin, 0) + term.clear_eos)
                do_describe_message_system(term, colors)
                waitprompt(term)
                dirty = 2
            elif inp.lower() == u'q':
                return
            if given_inp:
                # clear out line editor prompt
                echo(colors['backlight'](u'\b \b'))


def prompt_recipient(term, msg, colors, public=True):
    """ Prompt for recipient of message. """
    xpos = max(0, (term.width // 2) - (80 // 2))
    echo(term.move_x(xpos) + term.clear_eos)
    echo(u'Enter handle{0}.\r\n'.format(
        u', empty to address to all' if public else u''))
    echo(term.move_x(xpos) + ':: ')
    inp = LineEditor(username_max_length, msg.recipient,
                     colors={'highlight': colors['backlight']}
                     ).read()

    if inp is None:
        echo(u''.join((term.move_x(xpos),
                       colors['highlight']('Canceled.'),
                       term.clear_eol)))
        term.inkey(1)
        return False

    elif not inp.strip():
        # empty, recipient is None
        msg.recipient = None
        if public:
            return True
        return False

    inp = inp.strip()

    # validate/find user
    userlist = list_users()
    if inp in userlist:
        # exact match,
        msg.recipient = inp
        echo(u'\r\n')
        return True

    # nearest match
    for match in difflib.get_close_matches(inp.strip(), userlist):
        echo(u''.join((
            term.move_x(xpos),
            u'{0} [yn]'.format(colors['highlight'](match)),
            term.clear_eol,
            u' ?\b\b')))
        while True:
            inp = term.inkey()
            if inp.code == term.KEY_ESCAPE:
                # escape/cancel
                return False
            elif inp.lower() == u'y':
                # accept match
                msg.recipient = match
                echo(u'\r\n')
                return True
            elif inp.lower() == u'n':
                # next match
                break

    echo(u''.join((term.move_x(xpos),
                   colors['highlight']('No match.'),
                   term.clear_eol)))
    term.inkey(1)
    return False


def prompt_subject(term, msg, colors):
    """ Prompt for subject of message. """
    xpos = max(0, (term.width // 2) - (80 // 2))
    echo(u''.join((term.move_x(xpos),
                   term.clear_eos,
                   u'Enter Subject.\r\n',
                   term.move_x(xpos),
                   u':: ')))
    inp = LineEditor(subject_max_length, msg.subject,
                     colors={'highlight': colors['backlight']}
                     ).read()

    if inp is None or not inp.strip():
        echo(u''.join((term.move_x(xpos),
                       colors['highlight']('Canceled.'),
                       term.clear_eol)))
        term.inkey(1)
        return False

    msg.subject = inp.strip()
    return True


def prompt_body(term, msg, colors):
    """ Prompt for and set 'body' of message by executing 'editor' script. """
    with term.fullscreen():
        content = gosub('editor', save_key=None,
                        continue_draft=msg.body)
    # set syncterm font, if any
    if term.kind.startswith('ansi'):
        echo(syncterm_setfont(syncterm_font))
    echo(term.move(term.height, 0) + term.normal + u'\r\n')
    if content and content.strip():
        msg.body = content
        return True
    xpos = max(0, (term.width // 2) - (80 // 2))
    echo(u''.join((term.move_x(xpos),
                   colors['highlight']('Message canceled.'),
                   term.clear_eol)))
    term.inkey(1)
    return False


def prompt_tags(session, term, msg, colors, public=True):
    xpos = max(0, (term.width // 2) - (80 // 2))

    # conditionally enforce tag moderation
    moderated = get_ini('msg', 'moderated_tags', getter='getboolean')
    tag_moderators = set(get_ini('msg', 'tag_moderators', split=True))

    # enforce 'public' tag
    if public and 'public' not in msg.tags:
        msg.tags.add('public')
    elif not public and 'public' in msg.tags:
        msg.tags.remove('public')

    # describe all available tags, as we oft want to do.
    do_describe_available_tags(term, colors)

    # and remind ourselves of the available network tags,
    description = get_network_tag_description(term, colors)
    if description:
        show_description(term=term, color=None, description=description)

    echo(u''.join((term.move_x(xpos),
                   term.clear_eos,
                   u'Enter tags, separated by commas.\r\n',
                   term.move_x(xpos),
                   u':: ')))

    all_tags = list_tags()

    while True:
        inp = LineEditor(subject_max_length, u', '.join(sorted(msg.tags)),
                         colors={'highlight': colors['backlight']}
                         ).read()
        if inp is None:
            echo(u''.join((term.move_x(xpos),
                           colors['highlight']('Message canceled.'),
                           term.clear_eol)))
            term.inkey(1)
            return False

        msg.tags = set(filter(None, set(map(unicode.strip, inp.split(',')))))
        if moderated and not (tag_moderators | session.user.groups):
            cannot_tag = [_tag for _tag in msg.tags if _tag not in all_tags]
            if cannot_tag:
                echo(u''.join((u'\r\n', term.move_x(xpos),
                               u', '.join((quote(tag, colors)
                                           for tag in cannot_tag)),
                               u': not allowed; this system is moderated.')))
                term.inkey(2)
                echo(term.move_up)
                map(msg.tags.remove, cannot_tag)
                continue

        return True


def do_send_message(session, term, msg, colors):
    xpos = max(0, (term.width // 2) - (80 // 2))
    msg.save()
    do_mark_as_read(session, [msg.idx])
    echo(u''.join((u'\r\n',
                   term.move_x(xpos),
                   colors['highlight']('message sent!'))))
    session.send_event('global', ('newmsg', msg.idx,))
    term.inkey(1)



""" news script for x/84. """
# std
import os
import time
import codecs
import logging

# local
from x84.bbs import getterminal, getsession, echo
from x84.bbs import syncterm_setfont, decode_pipe
from common import display_banner, prompt_pager

#: filepath to folder containing this script
here = os.path.dirname(__file__)

#: filepath to news contents persisted to disk
news_file = os.path.join(os.path.dirname(__file__), 'art', 'news.txt')

#: encoding of news_file
news_file_encoding = 'utf8'

#: cached global memory view of news.txt
news_contents = None

#: filepath to artfile displayed for this script
art_file = os.path.join(here, 'art', 'news.ans')

#: encoding used to display artfile
art_encoding = 'cp437'

#: fontset for SyncTerm emulator
syncterm_font = 'topaz'

#: estimated art height (top of pager)
art_height = 8

log = logging.getLogger(__name__)


def main(quick=False):
    """
    Script entry point.

    :param bool quick: When True, returns early if this news has already
                       been read.
    """
    session, term = getsession(), getterminal()

    if not os.path.exists(news_file):
        log.warn('No news file, {0}'.format(news_file))
        echo(u'\r\n\r\n' + term.center(u'No news.').rstrip() + u'\r\n')
        return

    # return early if 'quick' is True and news is not new
    news_mtime = os.stat(news_file).st_mtime
    if quick and news_mtime < session.user.get('news_lastread', 0):
        return

    # set syncterm font, if any
    if syncterm_font and term.kind == 'ansi':
        echo(syncterm_setfont(syncterm_font))

    session.activity = 'Reading news'

    # display banner
    line_no = display_banner(filepattern=art_file, encoding=art_encoding)

    # retrieve news_file contents (decoded as utf8)
    news = decode_pipe(codecs.open(
        news_file, 'rb', news_file_encoding).read()
    ).splitlines()
    echo(u'\r\n\r\n')

    # display file contents, decoded, using a command-prompt pager.
    prompt_pager(content=news,
                 line_no=line_no + 2,
                 colors={'highlight': term.yellow,
                         'lowlight': term.green,
                         },
                 width=min(80, term.width))

    # update user's last-read time of news.
    session.user['news_lastread'] = time.time()
